---------------------------------------------------------------------------------------------------------------
-- RISA Project
-- Author: A.Greensted
-- Module:
-- Description:
-- History:
---------------------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library CommonLib;
use CommonLib.CommonPkg.all;

entity Fifo is
	generic(	dataWidth	: in		positive := 8);

	port (	clk			: in		std_logic;
				nReset		: in		std_logic;
				enable		: in		std_logic;

				wEn			: in		std_logic;
				rEn			: in		std_logic;

				dIn			: in		std_logic_vector(dataWidth-1 downto 0);
				dOut			: out		std_logic_vector(dataWidth-1 downto 0);
				
				full			: out		std_logic;
				empty			: out		std_logic;
				fillLevel	: out		std_logic_vector(1 downto 0));
end Fifo;

architecture General of Fifo is

	subtype WORD_TYPE is std_logic_vector(dataWidth-1 downto 0);
	type FIFO_RAM_TYPE is array (3 downto 1) of WORD_TYPE;

	signal fifoRam			: FIFO_RAM_TYPE;
	
	signal fullFlag		: std_logic;
	signal emptyFlag		: std_logic;
	signal readPointer	: unsigned(1 downto 0);

begin

fillLevel	<= std_logic_vector(readPointer);
full			<= fullFlag;
empty			<= emptyFlag;

OutputMux : process(readPointer, fifoRam, dIn)
begin
	
	case readPointer is
		when b"11" =>	dOut			<= fifoRam(3);
							fullFlag		<= '1';
							emptyFlag	<= '0';

		when b"10" =>	dOut			<= fifoRam(2);
							fullFlag		<= '0';
							emptyFlag	<= '0';

		when b"01" =>	dOut			<= fifoRam(1);
							fullFlag		<= '0';
							emptyFlag	<= '0';

		when others =>	dOut			<= (others => '0');
							fullFlag		<= '0';
							emptyFlag	<= '1';
	end case;

end process OutputMux;


FifoControl : process(clk)
begin
	if (clk'event and clk='1') then

		if (nReset='0') then
			readPointer	<= (others => '0');

		elsif (enable='1') then

			-- Write Only
			if (wEn='1' and rEn='0') then

				if (fullFlag='0') then -- Only allow write if not full
					readPointer <= readPointer + 1;
					fifoRam(3) <= fifoRam(2);
					fifoRam(2) <= fifoRam(1);
					fifoRam(1) <= dIn;
				end if;

			-- Read Only
			elsif (wEn='0' and rEn='1') then

				if (emptyFlag='0') then	-- Only allow read if not empty
					readPointer <= readPointer - 1;
				end if;

			-- Write and Read
			elsif (wEn='1' and rEn='1') then

				if (emptyFlag='1') then	-- If empty, increment pointer to null read attempt
					readPointer <= readPointer + 1;
				end if;

				fifoRam(3) <= fifoRam(2);
				fifoRam(2) <= fifoRam(1);
				fifoRam(1) <= dIn;

			end if;
		end if;
	end if;

end process;


end General;
