---------------------------------------------------------------------------------------------------------------
-- RISA Project
-- Author: A.Greensted
-- Module:
-- Description:
-- History:
---------------------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library CommonLib;
use CommonLib.CommonPkg.all;

library ConfigLib;
use ConfigLib.ConfigPkg.all;

entity ConfigAccessPort_CAP is
	port (	cap_clk				: in		std_logic;
				cap_nReset			: in		std_logic;
				cap_enable			: in		std_logic;
				cap_ctrlDataSel	: in		std_logic;
				cap_shiftExecSel	: in		std_logic;
				cap_dIn				: in		std_logic;
				cap_dOut				: out		std_logic;
				cap_busy				: out		std_logic;
				
				start					: out		std_logic;
				release				: in		std_logic;

				controlData			: out		std_logic_vector(10 downto 0);
				sendData				: out		std_logic_vector(15 downto 0);
				returnData			: in		std_logic_vector(15 downto 0));

end ConfigAccessPort_CAP;

architecture General of ConfigAccessPort_CAP is

	signal cap_start					: std_logic;
	signal cap_release				: std_logic;
	signal cap_dShiftEn				: std_logic;

	signal controlDataReg			: std_logic_vector(10 downto 0);
	signal sendDataReg				: std_logic_vector(15 downto 0);
	signal returnDataReg				: std_logic_vector(15 downto 0);

begin

cap_dShiftEn <= not cap_start and cap_enable and cap_shiftExecSel and not cap_ctrlDataSel;

-- StartReg -------------------------------------------------------------------------------------------------
-- Starts the Cross clock domain and configuration process
-- Once enabled, stops any access to the cap shift registers
----------------------------------------------------------------------------------------------------------------
StartReg : process(cap_clk)
begin

	if (cap_clk'event and cap_clk='1') then

		if (cap_nReset='0') then
			cap_start <= '0';

		elsif (cap_release='1') then
			cap_start <= '0';

		elsif (cap_enable='1' and cap_shiftExecSel='0') then
			cap_start <='1';

		end if;

	end if;

end process;

cap_busy <= cap_start;
start <= cap_start;

-- CapControlShiftReg -----------------------------------------------------------------------------------------
-- CAP Control shift register
----------------------------------------------------------------------------------------------------------------
CapControlShiftReg : process(cap_clk)
begin

	if (cap_clk'event and cap_clk='1') then

		if (cap_nReset='0') then
			controlDataReg <= (others => '0');

		elsif (cap_start='0' and cap_enable='1' and cap_shiftExecSel='1' and cap_ctrlDataSel='1') then
			controlDataReg <= controlDataReg(9 downto 0) & cap_dIn;

		end if;

	end if;

end process;

controlData <= controlDataReg;

-- CapSendShiftReg --------------------------------------------------------------------------------------------
-- CAP Data shift register, shifting config data in to chip
----------------------------------------------------------------------------------------------------------------
CapSendShiftReg : process(cap_clk)
begin

	if (cap_clk'event and cap_clk='1') then

		if (cap_nReset='0') then
			sendDataReg <= (others => '0');

		elsif (cap_dShiftEn='1') then
			sendDataReg <= sendDataReg(14 downto 0) & cap_dIn;

		end if;

	end if;

end process;

sendData <= sendDataReg;


-- CapSync ----------------------------------------------------------------------------------------------------
-- Synchronises the release signal to the cap clock domain
----------------------------------------------------------------------------------------------------------------
SYNC_CapSync : PulseSynchroniser
port map (	clk			=> cap_clk,
				dIn			=> release,
				dOut			=> cap_release);


-- CapReturnShiftReg ------------------------------------------------------------------------------------------
-- CAP Data shift register, shifting config data out of chip
----------------------------------------------------------------------------------------------------------------
CapReturnShiftReg : process(cap_clk)
begin

	if (cap_clk'event and cap_clk='1') then

		if (cap_nReset='0') then
			returnDataReg <= (others => '0');

		elsif (cap_release='1') then
			returnDataReg <= returnData;

		elsif (cap_dShiftEn='1') then
			returnDataReg <= returnDataReg(14 downto 0) & '0';

		end if;

	end if;

end process;

cap_dOut <= returnDataReg(15);

end General;
