---------------------------------------------------------------------------------------------------------------
-- RISA Project
-- Author: A.Greensted
-- Module:
-- Description:
-- History:
---------------------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library CommonLib;
use CommonLib.CommonPkg.all;

library ConfigLib;
use ConfigLib.ConfigPkg.all;

entity ConfigAccessPort_CNF is
	port (	clk					: in		std_logic;
				cnf_nReset			: in		std_logic;
				cnf_dIn				: in		std_logic;
				cnf_dOut				: out		std_logic;
				cnf_shiftEn			: out		std_logic;
				cnf_ctrl				: out		std_logic_vector(1 downto 0);
				cnf_target			: out		CONFIGTYPE_CAP_TARGET;
				cnf_targetSetEn	: out		std_logic_vector(3 downto 0);
				
				start					: in		std_logic;
				release				: out		std_logic;

				controlData			: in		std_logic_vector(10 downto 0);
				sendData				: in		std_logic_vector(15 downto 0);
				returnData			: out		std_logic_vector(15 downto 0));

end ConfigAccessPort_CNF;

architecture General of ConfigAccessPort_CNF is

	signal cnf_start					: std_logic;

	signal cnf_bitCount				: std_logic_vector(3 downto 0);
	signal cnf_ctrlTarget			: CONFIGTYPE_CAP_TARGET;
	signal cnf_ctrlMode				: std_logic_vector(1 downto 0);
	signal cnf_ctrlCell				: std_logic_vector(1 downto 0);

	signal cnf_memBitCount			: std_logic_vector(3 downto 0);

	signal sendDataReg				: std_logic_vector(15 downto 0);
	signal returnDataReg				: std_logic_vector(15 downto 0);

	signal toggle						: std_logic;

	signal cnf_dataOut				: std_logic;

	signal cnf_bitDecEn				: std_logic;
	signal cnf_releaseEn				: std_logic;
	signal cnf_sendShiftEn			: std_logic;
	signal cnf_returnShiftEn		: std_logic;
	signal cnf_dataFeedbackEn		: std_logic;
	signal cnf_memBitIncEn			: std_logic;
	signal cnf_memBitCountReset	: std_logic;

	type CNF_STATE_TYPE is (IDLE, TARGET_SET, DECODE, CTRL_SET, MEM_HOLD, MEM_FINISH);
	signal cS							: CNF_STATE_TYPE;
	signal nS							: CNF_STATE_TYPE; 

begin

-- CnfSync ----------------------------------------------------------------------------------------------------
-- Synchronises the 'release edge' to a pulse in the cap clock domain
----------------------------------------------------------------------------------------------------------------
SYNC_CnfSync : EdgeSynchroniser
port map (	clk	=> clk,
				dIn	=> start,
				dOut	=> cnf_start);


-- CnfBitCounter ----------------------------------------------------------------------------------------------
-- Counts out bits as they are shifted into the configuration chains
----------------------------------------------------------------------------------------------------------------
CnfBitCounter : process(clk)

	variable count : unsigned(3 downto 0);

begin
	if (clk'event and clk='1') then

		if (cnf_nReset='0') then
			count := (others => '0');

		elsif (cnf_start='1') then
			count := unsigned(controlData(3 downto 0));

		elsif (cnf_bitDecEn='1') then
			count := count - 1;

		end if;

	end if;

	cnf_bitCount <= std_logic_vector(count);

end process;

-- MemBitCounter ----------------------------------------------------------------------------------------------
-- Counts bits for shifting data into and out of memory target
----------------------------------------------------------------------------------------------------------------
MemBitCounter : process(clk)

	variable count : unsigned(3 downto 0);

begin
	if (clk'event and clk='1') then

		if (cnf_nReset='0') then
			count := (others => '0');

		elsif (cnf_memBitCountReset='1') then
			count := "0000";

		elsif (cnf_memBitIncEn='1') then
			count := count + 1;

		end if;

	end if;

	cnf_memBitCount <= std_logic_vector(count);

end process;


-- CnfCtrlReg -------------------------------------------------------------------------------------------------
-- Configuration control registers
----------------------------------------------------------------------------------------------------------------
CnfControlReg : process(clk)
begin

	if (clk'event and clk='1') then

		if (cnf_nReset='0') then
			cnf_ctrlCell	<= (others => '0');
			cnf_ctrlTarget	<= CONFIG_CAP_TARGET_NULL;
			cnf_ctrlMode	<= (others => '0');

		elsif (cnf_start='1') then
			cnf_ctrlCell	<= controlData(10 downto 9);
			cnf_ctrlTarget	<= to_CONFIGTYPE_CAP_TARGET(controlData(8 downto 6));
			cnf_ctrlMode	<= controlData(5 downto 4);

		end if;

	end if;

end process;

cnf_target <= cnf_ctrlTarget;


-- CapSendShiftReg --------------------------------------------------------------------------------------------
-- CAP Data shift register, shifting config data into chip
----------------------------------------------------------------------------------------------------------------
CnfSendShiftReg : process(clk)
begin

	if (clk'event and clk='1') then

		if (cnf_nReset='0') then
			sendDataReg <= (others => '0');

		elsif (cnf_start='1') then
			sendDataReg <= sendData;

		elsif (cnf_sendShiftEn='1') then
			sendDataReg <= sendDataReg(14 downto 0) & '0';

		end if;

	end if;

end process;

cnf_dataOut <= sendDataReg(15);
cnf_dOut <= cnf_dataOut;

-- ReleaseToggle ----------------------------------------------------------------------------------------------
-- Provides the toggle part of the release pusle synchroniser
----------------------------------------------------------------------------------------------------------------
ReleaseToggle : process(clk)
begin

	if (clk'event and clk='1') then

		if (cnf_nReset='0') then
			toggle <= '0';

		else
			if (cnf_releaseEn='1') then
				toggle <= not toggle;
			else
				toggle <= toggle;
			end if;

		end if;

	end if;

end process;

release <= toggle;

-- CnfReturnShiftReg ------------------------------------------------------------------------------------------
-- CAP data shift register - shifting config data out of chip
----------------------------------------------------------------------------------------------------------------
CnfReturnShiftReg : process(clk)
	variable shiftData : std_logic;
begin

	if (clk'event and clk='1') then

		if (cnf_nReset='0') then
			returnDataReg <= (others => '0');

		elsif (cnf_returnShiftEn='1') then

			if (cnf_dataFeedbackEn='1') then
				shiftData := cnf_dataOut;
			else
				shiftData := cnf_dIn;
			end if;

			returnDataReg <= returnDataReg(14 downto 0) & shiftData;

		end if;

	end if;

end process;

returnData <= returnDataReg;

-- CnfFSM -----------------------------------------------------------------------------------------------------
-- Configuration Finite State Machine
----------------------------------------------------------------------------------------------------------------
CnfFSM_Control : process (clk)
begin

	if (clk'event and clk='1') then

		if (cnf_nReset='0') then
			cS <= IDLE;
		else
			cS <= nS;
		end if;

	end if;
end process;

CnfFSM_Logic : process(cS, nS, cnf_start, cnf_bitCount, cnf_ctrlCell, cnf_ctrlTarget,
								cnf_ctrlMode, cnf_memBitCount)
begin

	-- Default Assignments
	cnf_targetSetEn		<= "0000";
	cnf_ctrl					<= "00";
	cnf_shiftEn				<= '0';
	cnf_sendShiftEn		<= '0';
	cnf_returnShiftEn		<= '0';
	cnf_dataFeedbackEn	<= '0';
	cnf_bitDecEn			<= '0';
	cnf_memBitIncEn		<= '0';
	cnf_memBitCountReset	<= '1';
	cnf_releaseEn			<= '0';
	nS							<= IDLE;

	case cS is
		-- IDLE
		-- The FSM sits in the IDLE state waiting for a pulse on cnf_start.
		-- The pulse starts the FSM, which jumps straigh to the TARGET SET state.
		-- This provides a one clock pulse pause for cnf_ctrlCell, cnf_ctrlTarget and cnf_ctrlMode
		-- to be set.
		-- {{{
		when IDLE =>

			if (cnf_start='1') then
				nS <= TARGET_SET;	-- Go to the TARGET_SET state
			else
				nS	<= IDLE;		-- Stay in the IDLE state
			end if;
		-- }}}

		-- TARGET SET
		-- This trigers the a pulse to the required cnf_targetSetEn line, which connect to
		-- a ConfigArbiter.
		-- {{{
		when TARGET_SET =>

			case cnf_ctrlCell is
				when "00"	=> cnf_targetSetEn <= "0001";
				when "01"	=> cnf_targetSetEn <= "0010";
				when "10"	=> cnf_targetSetEn <= "0100";
				when "11"	=> cnf_targetSetEn <= "1000";
				when others	=>
					assert false
					report "ConfAccessPort.CnfFSM_Logic: Bad cnf_ctrlCell value"
					severity failure;
			end case;

			nS <= DECODE;		-- Go to the DECODE state
		-- }}}

		-- DECODE
		-- The DECODE state. Depending on the requested configuration target, this will
		-- shift data and set the cnf_ctrl lines
		-- {{{
		when DECODE =>

			case cnf_ctrlTarget is

				--== NULL ==--
				when CONFIG_CAP_TARGET_NULL =>

					case cnf_ctrlMode is

						-- FEEDBACK --
						when "01" =>
							cnf_dataFeedbackEn	<= '1';		-- Feedback the cnf data
							cnf_sendShiftEn		<= '1';		-- Shift from Send reg
							cnf_returnShiftEn		<= '1';		-- Shift into return reg
							cnf_bitDecEn			<= '1';		-- Decrement the bit count

							if (cnf_bitCount=b"0000") then
								-- Once all bits have been shifted
								cnf_releaseEn	<= '1';		-- Release the conf unit
								nS					<= IDLE;		-- Return to the IDLE state
							else
								-- Keep Shifting
								nS <= DECODE;		-- Stay in the DECODE state
							end if;

						when others =>
							-- Any other mode
							cnf_releaseEn	<= '1';		-- Release the conf unit
							nS					<= IDLE;		-- Return to IDLE state

					end case;

				--== UP CONTROL ==--
				when CONFIG_CAP_TARGET_UP_CONTROL =>

					cnf_shiftEn			<= '1';		-- Enable shift
					cnf_sendShiftEn	<= '1';		-- Shift from send reg
					cnf_returnShiftEn	<= '1';		-- Shift into return reg
					cnf_bitDecEn		<= '1';		-- Decrement the bit count

					if (cnf_bitCount=b"0000") then
						-- Once all bits have been shifted
						nS <= CTRL_SET;	-- Go to the CTRL_SET state
					else
						-- Keep Shifting
						nS <= DECODE;		-- Stay in the DECODE state
					end if;


				--== LOGIC, ROUTE ==--
				when CONFIG_CAP_TARGET_LOGIC_CNF_CHAIN | CONFIG_CAP_TARGET_ROUTE_CNF_CHAIN =>

					case cnf_ctrlMode is

						-- SHIFT --
						when "00" =>
							cnf_shiftEn			<= '1';		-- Enable shift
							cnf_sendShiftEn	<= '1';		-- Shift from send reg
							cnf_returnShiftEn	<= '1';		-- Shift into return reg
							cnf_bitDecEn		<= '1';		-- Decrement the bit count

							if (cnf_bitCount=b"0000") then
								-- Once all bits have been shifted
								cnf_releaseEn	<= '1';		-- Release the conf unit
								nS					<= IDLE;		-- Return to the IDLE state
							else
								-- Keep Shifting
								nS <= DECODE;		-- Stay in the DECODE state
							end if;

						-- LOAD --
						when "01" =>
							cnf_ctrl			<= "01";		-- Set loadEn
							cnf_releaseEn	<= '1';		-- Release the conf unit
							nS					<= IDLE;		-- Return to IDLE state

						-- READBACK --
						when "10" =>
							cnf_ctrl			<= "10";		-- Set readbackEn
							cnf_releaseEn	<= '1';		-- Release the conf unit
							nS					<= IDLE;		-- Return to IDLE state

						when others =>
							cnf_releaseEn	<= '1';		-- Release the conf unit
							nS					<= IDLE;		-- Return to IDLE state

					end case;


				--== LOGIC SELECT, ROUTE SELECT ==--
				when CONFIG_CAP_TARGET_LOGIC_SEL_CHAIN | CONFIG_CAP_TARGET_ROUTE_SEL_CHAIN =>

					case cnf_ctrlMode is

						-- SHIFT --
						when "00" =>

							cnf_shiftEn			<= '1';		-- Keep shifting
							cnf_sendShiftEn	<= '1';		-- Shift from Send reg
							cnf_returnShiftEn	<= '1';		-- Shift into return reg
							cnf_bitDecEn		<= '1';		-- Decrement the bit count

							if (cnf_bitCount=b"0000") then
								-- Once all bits have been shifted
								cnf_releaseEn	<= '1';		-- Release the conf unit
								nS					<= IDLE;		-- Return to the IDLE state
							else
								-- Keep Shifting
								nS <= DECODE;		-- Stay in the DECODE state
							end if;

						-- SHIFT FEEDBACK --
						when "01" =>

							cnf_ctrl				<= "01";		-- Set feedbackEn
							cnf_shiftEn			<= '1';		-- Keep shifting
							cnf_sendShiftEn	<= '1';		-- Shift from Send reg
							cnf_returnShiftEn	<= '1';		-- Shift into return reg
							cnf_bitDecEn		<= '1';		-- Decrement the bit count

							if (cnf_bitCount=b"0000") then
								-- Once all bits have been shifted
								cnf_releaseEn	<= '1';		-- Release the conf unit
								nS					<= IDLE;		-- Return to the IDLE state
							else
								-- Keep Shifting
								nS					<= DECODE;		-- Stay in the DECODE state
							end if;

						when others =>
							cnf_releaseEn	<= '1';		-- Release the conf unit
							nS					<= IDLE;		-- Return to IDLE state

					end case;

				--== MEM DATA ==--
				when CONFIG_CAP_TARGET_UP_MEM_DATA =>
				
					case cnf_ctrlMode is

						-- READ --
						when "00" =>
							cnf_ctrl	<= "01";		-- Set readEn
							nS			<= MEM_HOLD;

						-- WRITE, READ & WRITE --
						when "01" | "10" =>
							cnf_shiftEn				<= '1';		-- Keep Shifting
							cnf_sendShiftEn		<= '1';		-- Shift from send reg only
							cnf_memBitIncEn		<= '1';		-- Increment the mem bit counter
							cnf_memBitCountReset	<= '0';		-- Allow counter to increment

							if (cnf_memBitCount=b"1111") then			-- Counts to 15 for 16 data bits
								-- Once all bits have been shifted
								nS <= CTRL_SET;		-- Go to the CTRL_SETstate
							else
								-- Keep Shifting
								nS <= DECODE;		-- Stay in the DECODE state
							end if;

						when others =>
							-- Any other mode
							cnf_releaseEn	<= '1';		-- Release the conf unit
							nS					<= IDLE;		-- Return to IDLE state

					end case;

				--== MEM ADDRESS ==--
				when CONFIG_CAP_TARGET_UP_MEM_ADDRESS =>

					cnf_shiftEn				<= '1';		-- Keep Shifting
					cnf_sendShiftEn		<= '1';		-- Shift from send reg only
					cnf_memBitIncEn		<= '1';		-- Increment the mem bit counter
					cnf_memBitCountReset	<= '0';		-- Allow counter to increment

					if (cnf_memBitCount=b"1100") then			-- Counts to 12 for 13 address bits
						-- Once all bits have been shifted
						cnf_releaseEn	<= '1';		-- Release the conf unit
						nS					<= IDLE;		-- Go to the IDLE state
					else
						-- Keep Shifting
						nS <= DECODE;		-- Stay in the DECODE state
					end if;

			end case;
			-- }}}

		-- CTRL_SET
		-- The CTRL_SET state sets the control lines after (and before) and shift
		-- has occured. Used for uP Control and Memory
		-- {{{
		when CTRL_SET =>

			case cnf_ctrlTarget is

				--== UP CONTROL ==--
				when CONFIG_CAP_TARGET_UP_CONTROL =>
					cnf_ctrl			<= "01";		-- Set ctrlSetEn
					cnf_releaseEn	<= '1';		-- Release the conf unit
					nS					<= IDLE;		-- Return to IDLE state

				--== MEM DATA  ==--
				when CONFIG_CAP_TARGET_UP_MEM_DATA =>

					-- WRITE --
					if (cnf_ctrlMode="01") then
						cnf_ctrl <= "10";		-- Set writeEn

					-- READ & WRITE --
					else
						cnf_ctrl <= "11";		-- Set readEn and writeEn

					end if;

					nS <= MEM_HOLD;	-- Got to the MEM_HOLD state

				when others =>
					cnf_releaseEn	<= '1';		-- Release the conf unit
					nS					<= IDLE;		-- Return to IDLE state

			end case;

		-- }}}

		-- MEM_HOLD
		-- The MEM_HOLD state provide a single clock pulse delay required by the
		-- Configurable Memory to complete a memory read or write.
		-- {{{
		when MEM_HOLD =>

			case cnf_ctrlMode is

				-- READ, READ & WRITE--
				when "00" | "10" =>
					nS <= MEM_FINISH;	-- Go to MEM_FINISH state

				-- WRITE and others
				when others =>
					cnf_releaseEn	<= '1';		-- Release the conf unit
					nS					<= IDLE;		-- return to IDLE state

			end case;
		-- }}}

		-- MEM_FINISH
		-- The MEM_FINISH state shifts memory data into the return reg
		-- when a memory read has been requested
		-- {{{
		when MEM_FINISH =>

			cnf_shiftEn				<= '1';		-- Keep shifting
			cnf_returnShiftEn		<= '1';		-- Shift into return reg
			cnf_memBitIncEn		<= '1';		-- Increment the mem bit counter
			cnf_memBitCountReset	<= '0';		-- Allow counter to increment

			if (cnf_memBitCount=b"1111") then
				-- Once all bits have been shifted
				cnf_releaseEn	<= '1';		-- Release the conf unit
				nS					<= IDLE;		-- Return to the IDLE state
			else
				-- Keep Shifting
				nS <= MEM_FINISH;		-- Stay in the MEM_FINISH state
			end if;

		-- }}}

	end case;

end process;

end General;
