---------------------------------------------------------------------------------------------------------------
-- RISA Project
-- Author: A.Greensted
-- Module:
-- Description:
-- History:
---------------------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

library ConfigLib;
use ConfigLib.ConfigPkg.all;

entity ConfigArbiter is
	port (	clk									: in		std_logic;
				nReset								: in		std_logic;

				cnf_dIn								: in		std_logic;
				cnf_dOut								: out		std_logic;
				cnf_shiftEn							: in		std_logic;
				cnf_ctrl								: in		std_logic_vector(1 downto 0);
				cnf_target							: in		CONFIGTYPE_CAP_TARGET;
				cnf_targetSetEn					: in		std_logic;

				-- Logic Configuration driving FPGA
				fpga_logicCnfChain_dIn			: in		std_logic;
				fpga_logicCnfChain_dOut			: out		std_logic;
				fpga_logicCnfChain_shiftEn		: out		std_logic;
				fpga_logicCnfChain_loadEn		: out		std_logic;
				fpga_logicCnfChain_readBackEn	: out		std_logic;

				-- Logic Configuration from uP
				uP_logicCnfChain_dIn				: in		std_logic;
				uP_logicCnfChain_dOut			: out		std_logic;
				uP_logicCnfChain_shiftEn		: in		std_logic;
				uP_logicCnfChain_loadEn			: in		std_logic;
				uP_logicCnfChain_readBackEn	: in		std_logic;

				uP_logicCnfChain_lockRequest	: in		std_logic;
				uP_logicCnfChain_lock			: out		std_logic;

				-- Logic Target Configuration driving fabric
				fpga_logicSelChain_dIn			: in		std_logic;
				fpga_logicSelChain_dOut			: out		std_logic;
				fpga_logicSelChain_shiftEn		: out		std_logic;

				-- Logic Target Configuration from uP
				uP_logicSelChain_dIn				: in		std_logic;
				uP_logicSelChain_dOut			: out		std_logic;
				uP_logicSelChain_shiftEn		: in		std_logic;
				uP_logicSelChain_feedbackEn	: in		std_logic;

				uP_logicSelChain_lockRequest	: in		std_logic;
				uP_logicSelChain_lock			: out		std_logic;

				-- Route Configuration driving Fabric
				fpga_routeCnfChain_dIn			: in		std_logic;
				fpga_routeCnfChain_dOut			: out		std_logic;
				fpga_routeCnfChain_shiftEn		: out		std_logic;
				fpga_routeCnfChain_loadEn		: out		std_logic;
				fpga_routeCnfChain_readBackEn	: out		std_logic;

				-- Route Configuration from uP
				uP_routeCnfChain_dIn				: in		std_logic;
				uP_routeCnfChain_dOut			: out		std_logic;
				uP_routeCnfChain_shiftEn		: in		std_logic;
				uP_routeCnfChain_loadEn			: in		std_logic;
				uP_routeCnfChain_readBackEn	: in		std_logic;

				uP_routeCnfChain_lockRequest	: in		std_logic;
				uP_routeCnfChain_lock			: out		std_logic;

				-- Route Target Configuration driving fabric
				fpga_routeSelChain_dIn			: in		std_logic;
				fpga_routeSelChain_dOut			: out		std_logic;
				fpga_routeSelChain_shiftEn		: out		std_logic;

				-- Route Target Configuration from uP
				uP_routeSelChain_dIn				: in		std_logic;
				uP_routeSelChain_dOut			: out		std_logic;
				uP_routeSelChain_shiftEn		: in		std_logic;
				uP_routeSelChain_feedbackEn	: in		std_logic;

				uP_routeSelChain_lockRequest	: in		std_logic;
				uP_routeSelChain_lock			: out		std_logic;

				-- Memory Configuration
				uP_memCnf_dIn						: in		std_logic;
				uP_memCnf_dOut						: out		std_logic;
				uP_memCnf_shiftEn					: out		std_logic;
				uP_memCnf_addDataSel				: out		std_logic;
				uP_memCnf_readEn					: out		std_logic;
				uP_memCnf_writeEn					: out		std_logic;

				-- uP Control
				uP_control_nReset					: out		std_logic;
				uP_control_enable					: out		std_logic);

end ConfigArbiter;

architecture General of ConfigArbiter is

	signal targetReg		: CONFIGTYPE_CAP_TARGET;
	signal uPControlReg	: std_logic_vector(1 downto 0);

begin

-- TargetLatch -----------------------------------------------------------------------------------------------------
-- Latches the desired target when cnf_targetSetEn is high
----------------------------------------------------------------------------------------------------------------
TargetLatch : process (clk)
begin

	if (clk'event and clk='1') then

		if (nReset='0') then
			targetReg <= CONFIG_CAP_TARGET_NULL;

		elsif (cnf_targetSetEn = '1') then
			targetReg <= cnf_target;

		end if;

	end if;

end process;

UPControl : process (clk)
begin
	if (clk'event and clk='1') then

		if (nReset='0') then
			uP_control_nReset	<= '1';
			uP_control_enable	<= '0';
			uPControlReg		<= b"00";

		elsif (targetReg=CONFIG_CAP_TARGET_UP_CONTROL) then

			if (cnf_ctrl(0)='1') then
				uP_control_nReset	<= uPControlReg(0);
				uP_control_enable	<= uPControlReg(1);

			elsif (cnf_shiftEn = '1') then
				uPControlReg <= uPControlReg(0) & cnf_dIn;

			end if;

		end if;

	end if;

end process;

-- LogicCnfChainSelector --------------------------------------------------------------------------------------
--
----------------------------------------------------------------------------------------------------------------
-- {{{
LogicCnfChainSelector : process (targetReg, cnf_dIn, cnf_shiftEn, cnf_ctrl,
											uP_logicCnfChain_dIn, uP_logicCnfChain_shiftEn,
											uP_logicCnfChain_loadEn, uP_logicCnfChain_readBackEn)
begin

	-- Logic Configuration Chain
	if (targetReg = CONFIG_CAP_TARGET_LOGIC_CNF_CHAIN) then
		fpga_logicCnfChain_dOut			<= cnf_dIn;
		fpga_logicCnfChain_shiftEn		<= cnf_shiftEn;
		fpga_logicCnfChain_loadEn		<= cnf_ctrl(0);
		fpga_logicCnfChain_readBackEn	<= cnf_ctrl(1);

	else
		fpga_logicCnfChain_dOut			<= uP_logicCnfChain_dIn;
		fpga_logicCnfChain_shiftEn		<= uP_logicCnfChain_shiftEn;
		fpga_logicCnfChain_loadEn		<= uP_logicCnfChain_loadEn;
		fpga_logicCnfChain_readBackEn	<= uP_logicCnfChain_readBackEn;

	end if;

end process; -- }}}

-- LogicSelChainSelector --------------------------------------------------------------------------------------
--
----------------------------------------------------------------------------------------------------------------
-- {{{
LogicSelChainSelector : process (targetReg, cnf_ctrl, fpga_logicSelChain_dIn, cnf_dIn, cnf_shiftEn,
											uP_logicSelChain_feedbackEn, uP_logicSelChain_dIn, uP_logicSelChain_shiftEn)

begin
	-- Logic Selection Chain
	if (targetReg = CONFIG_CAP_TARGET_LOGIC_SEL_CHAIN) then
		if (cnf_ctrl(0) = '1') then
			fpga_logicSelChain_dOut	<= fpga_logicSelChain_dIn;
		else
			fpga_logicSelChain_dOut	<= cnf_dIn;
		end if;

		fpga_logicSelChain_shiftEn	<= cnf_shiftEn;

	else
		if (uP_logicSelChain_feedbackEn = '1') then
			fpga_logicSelChain_dOut	<= fpga_logicSelChain_dIn;
		else
			fpga_logicSelChain_dOut	<= uP_logicSelChain_dIn;
		end if;

		fpga_logicSelChain_shiftEn	<= uP_logicSelChain_shiftEn;

	end if;

end process; -- }}}

-- RouteCnfChainSelector --------------------------------------------------------------------------------------
--
----------------------------------------------------------------------------------------------------------------
-- {{{
RouteCnfChainSelector : process (targetReg, cnf_dIn, cnf_shiftEn, cnf_ctrl,
											uP_routeCnfChain_dIn, uP_routeCnfChain_shiftEn,
											uP_routeCnfChain_loadEn, uP_routeCnfChain_readBackEn)
begin

	-- Route Configuration Chain
	if (targetReg = CONFIG_CAP_TARGET_ROUTE_CNF_CHAIN) then
		fpga_routeCnfChain_dOut			<= cnf_dIn;
		fpga_routeCnfChain_shiftEn		<= cnf_shiftEn;
		fpga_routeCnfChain_loadEn		<= cnf_ctrl(0);
		fpga_routeCnfChain_readBackEn	<= cnf_ctrl(1);

	else
		fpga_routeCnfChain_dOut			<= uP_routeCnfChain_dIn;
		fpga_routeCnfChain_shiftEn		<= uP_routeCnfChain_shiftEn;
		fpga_routeCnfChain_loadEn		<= uP_routeCnfChain_loadEn;
		fpga_routeCnfChain_readBackEn	<= uP_routeCnfChain_readBackEn;

	end if;

end process; -- }}}

-- RouteSelChainSelector --------------------------------------------------------------------------------------
--
----------------------------------------------------------------------------------------------------------------
-- {{{
RouteSelChainSelector : process (targetReg, cnf_ctrl, fpga_routeSelChain_dIn, cnf_dIn, cnf_shiftEn,
											uP_routeSelChain_feedbackEn, uP_routeSelChain_dIn, uP_routeSelChain_shiftEn)

begin
	-- Route Selection Chain
	if (targetReg = CONFIG_CAP_TARGET_ROUTE_SEL_CHAIN) then
		if (cnf_ctrl(0) = '1') then
			fpga_routeSelChain_dOut	<= fpga_routeSelChain_dIn;
		else
			fpga_routeSelChain_dOut	<= cnf_dIn;
		end if;

		fpga_routeSelChain_shiftEn	<= cnf_shiftEn;

	else
		if (uP_routeSelChain_feedbackEn = '1') then
			fpga_routeSelChain_dOut	<= fpga_routeSelChain_dIn;
		else
			fpga_routeSelChain_dOut	<= uP_routeSelChain_dIn;
		end if;

		fpga_routeSelChain_shiftEn	<= uP_routeSelChain_shiftEn;

	end if;

end process; --}}}

-- MemCnfSelector ---------------------------------------------------------------------------------------------
--
----------------------------------------------------------------------------------------------------------------
-- {{{
MemCnfSelector : process (targetReg, cnf_dIn, cnf_shiftEn, cnf_ctrl)
begin

	if (targetReg = CONFIG_CAP_TARGET_UP_MEM_ADDRESS) then
		uP_memCnf_dOut			<= cnf_dIn;
		uP_memCnf_shiftEn		<= cnf_shiftEn;
		uP_memCnf_addDataSel	<= '1';
		uP_memCnf_readEn		<= cnf_ctrl(0);
		uP_memCnf_writeEn		<= cnf_ctrl(1);

	elsif (targetReg = CONFIG_CAP_TARGET_UP_MEM_DATA) then
		uP_memCnf_dOut			<= cnf_dIn;
		uP_memCnf_shiftEn		<= cnf_shiftEn;
		uP_memCnf_addDataSel	<= '0';
		uP_memCnf_readEn		<= cnf_ctrl(0);
		uP_memCnf_writeEn		<= cnf_ctrl(1);

	else
		uP_memCnf_dOut			<= '0';
		uP_memCnf_shiftEn		<= '0';
		uP_memCnf_addDataSel	<= '0';
		uP_memCnf_readEn		<= '0';
		uP_memCnf_writeEn		<= '0';

	end if;

end process; -- }}}

OutputSelector : process(	targetReg, uPControlReg,
									fpga_logicCnfChain_dIn, fpga_logicSelChain_dIn,
									fpga_routeCnfChain_dIn,	fpga_routeSelChain_dIn,
									uP_memCnf_dIn)
begin

	case targetReg is
		when CONFIG_CAP_TARGET_UP_CONTROL =>
			cnf_dOut <= uPControlReg(1);

		when CONFIG_CAP_TARGET_LOGIC_CNF_CHAIN =>
			cnf_dOut <= fpga_logicCnfChain_dIn;

		when CONFIG_CAP_TARGET_LOGIC_SEL_CHAIN =>
			cnf_dOut <= fpga_logicSelChain_dIn;

		when CONFIG_CAP_TARGET_ROUTE_CNF_CHAIN =>
			cnf_dOut <= fpga_routeCnfChain_dIn;

		when CONFIG_CAP_TARGET_ROUTE_SEL_CHAIN =>
			cnf_dOut <= fpga_routeSelChain_dIn;

		when CONFIG_CAP_TARGET_UP_MEM_DATA | CONFIG_CAP_TARGET_UP_MEM_ADDRESS =>
			cnf_dOut <= uP_memCnf_dIn;

		when others =>
			cnf_dOut <= '0';

	end case;

end process;

uP_logicCnfChain_dOut <= fpga_logicCnfChain_dIn;
uP_logicSelChain_dOut <= fpga_logicSelChain_dIn;
uP_routeCnfChain_dOut <= fpga_routeCnfChain_dIn;
uP_routeSelChain_dOut <= fpga_routeSelChain_dIn;

-- Lock Bits --------------------------------------------------------------------------------------------------
-- Controls the Chain Lock bits
----------------------------------------------------------------------------------------------------------------
-- {{{
LockBits : process(clk)
begin

	if (clk'event and clk='1') then

		if (nReset='0') then
			uP_logicCnfChain_lock <= '0';
			uP_logicSelChain_lock <= '0';
			uP_routeCnfChain_lock <= '0';
			uP_routeSelChain_lock <= '0';

		else

			-- Logic Configuration Chain
			if (cnf_target=CONFIG_CAP_TARGET_LOGIC_CNF_CHAIN and cnf_targetSetEn='1') then
				uP_logicCnfChain_lock <= '0';

			elsif (targetReg=CONFIG_CAP_TARGET_LOGIC_CNF_CHAIN) then
				uP_logicCnfChain_lock <= '0';

			elsif (uP_logicCnfChain_lockRequest='1') then
				uP_logicCnfChain_lock <= '1';

			end if;

			-- Logic Selection Chain
			if (cnf_target=CONFIG_CAP_TARGET_LOGIC_SEL_CHAIN and cnf_targetSetEn='1') then
				uP_logicSelChain_lock <= '0';

			elsif (targetReg=CONFIG_CAP_TARGET_LOGIC_SEL_CHAIN) then
				uP_logicSelChain_lock <= '0';

			elsif (uP_logicSelChain_lockRequest='1') then
				uP_logicSelChain_lock <= '1';

			end if;

			-- Route Configuration Chain
			if (cnf_target=CONFIG_CAP_TARGET_ROUTE_CNF_CHAIN and cnf_targetSetEn='1') then
				uP_routeCnfChain_lock <= '0';

			elsif (targetReg=CONFIG_CAP_TARGET_ROUTE_CNF_CHAIN) then
				uP_routeCnfChain_lock <= '0';

			elsif (uP_routeCnfChain_lockRequest='1') then
				uP_routeCnfChain_lock <= '1';

			end if;

			-- Route Selection Chain
			if (cnf_target=CONFIG_CAP_TARGET_ROUTE_SEL_CHAIN and cnf_targetSetEn='1') then
				uP_routeSelChain_lock <= '0';

			elsif (targetReg=CONFIG_CAP_TARGET_ROUTE_SEL_CHAIN) then
				uP_routeSelChain_lock <= '0';

			elsif (uP_routeSelChain_lockRequest='1') then
				uP_routeSelChain_lock <= '1';

			end if;

		end if;

	end if;

end process; -- }}}

end General;
