---------------------------------------------------------------------------------------------------------------
-- RISA Project
-- Author: A.Greensted
-- Module:
-- Description:
-- History:
---------------------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library CommonLib;
use CommonLib.CommonPKG.all;

entity ConfigurableMemory is
	port (	clk				: in		std_logic;
				enable			: in		std_logic;
				nReset			: in		std_logic;

				wEn				: in		std_logic;
				add				: in		std_logic_vector(12 downto 0);
				dIn				: in		std_logic_vector(15 downto 0);

				dOutLU			: out		std_logic_vector(15 downto 0);
				dOutU				: out		std_logic_vector(15 downto 0);
				
				-- Configuration
				cnf_dIn			: in		std_logic;
				cnf_dOut			: out		std_logic;
				cnf_shiftEn		: in		std_logic;
				cnf_addDataSel	: in		std_logic;
				cnf_readEn		: in		std_logic;
				cnf_writeEn		: in		std_logic);
end ConfigurableMemory;

architecture General of ConfigurableMemory is

	constant	DATA_SEL			: std_logic := '0';
	constant	ADD_SEL			: std_logic := '1';

	-- Memory Signals
	signal mem_en				: std_logic;
	signal mem_wEn				: std_logic;
	signal mem_add				: std_logic_vector(12 downto 0);
	signal mem_dIn				: std_logic_vector(15 downto 0);
	signal mem_dOutLU			: std_logic_vector(15 downto 0);
	signal mem_dOutU			: std_logic_vector(15 downto 0);

	-- Configuration Signals
	signal cnf_readEn2		: std_logic;
	signal cnf_writeEn2		: std_logic;

	signal cnf_en				: std_logic;
	signal cnf_addIncEn		: std_logic;
	signal cnf_memWEn			: std_logic;
	signal cnf_memEn			: std_logic;
	signal cnf_dOutLoadEn	: std_logic;

	signal cnf_dInReg			: std_logic_vector(15 downto 0);
	signal cnf_addReg			: std_logic_vector(12 downto 0);
	signal cnf_dOutReg		: std_logic_vector(15 downto 0);

	type CNF_STAGE is (STAGE1, STAGE2);
	signal cS					: CNF_STAGE;
	signal nS					: CNF_STAGE;

begin

-- ReadWriteDelay ---------------------------------------------------------------------------------------------
-- Single clock cycle delay for readEn and writeEn, used by Stage 2 of the FSM
----------------------------------------------------------------------------------------------------------------
ReadWriteDelay : process (clk)
begin
	
	if (clk'event and clk='1') then

		if (nReset='0') then
			cnf_readEn2		<= '0';
			cnf_writeEn2	<= '0';

		else
			cnf_readEn2		<= cnf_readEn;
			cnf_writeEn2	<= cnf_writeEn;

		end if;

	end if;

end process;


-- CnfFSM ------------------------------------------------------------------------------------------------------
-- Configuration Finite State Machine
-- Stage 1, loads Memory dOut (from address held in address shift reg) to the output shift register
-- Stage 2, writes data from input data shift register, to address, then increments address
----------------------------------------------------------------------------------------------------------------
CnfFSM_Control : process (clk)
begin
	if (clk'event and clk='1') then

		if (nReset='0') then
			cS <= STAGE1;

		else
			cS <= nS;

		end if;

	end if;

end process;

CnfFSM_Logic : process(cS, nS, cnf_writeEn, cnf_readEn, cnf_writeEn2, cnf_readEn2)
begin

	case cS is

		when STAGE1 =>

			if (cnf_writeEn='1' or cnf_readEn='1') then
				cnf_en <= '1';

				if (cnf_readEn = '1') then
					cnf_memEn <= '1';
				else
					cnf_memEn <= '0';
				end if;

				nS <= STAGE2;

			else
				cnf_en		<= '0';
				cnf_memEn	<= '0';
				nS				<= STAGE1;

			end if;

			cnf_memWEn		<= '0';
			cnf_dOutLoadEn	<= '0';
			cnf_addIncEn	<= '0';
				
		when STAGE2 =>
			cnf_en		<= '1';
			cnf_memEn	<= '1';
			nS				<= STAGE1;

			if (cnf_writeEn2='1') then
				cnf_memWEn <= '1';
			else 
				cnf_memWEn <= '0';
			end if;

			if (cnf_readEn2='1') then
				cnf_dOutLoadEn <= '1';
			else 
				cnf_dOutLoadEn <= '0';
			end if;

			cnf_addIncEn	<= '1';

	end case;

end process;


-- DInShiftReg -------------------------------------------------------------------------------------------------
-- Data input shift register
----------------------------------------------------------------------------------------------------------------
DInShiftReg : process(clk)
begin
	if (clk'event and clk='1') then

		if (nReset='0') then
			cnf_dInReg <= (others => '0');

		elsif (cnf_addDataSel='0' and cnf_en='0' and cnf_shiftEn='1') then
			cnf_dInReg <= cnf_dInReg(14 downto 0) & cnf_dIn;

		end if;

	end if;

end process;

-- AddShiftReg -------------------------------------------------------------------------------------------------
-- Address shift register
----------------------------------------------------------------------------------------------------------------
AddShiftReg : process(clk)
begin
	if (clk'event and clk='1') then

		if (nReset='0') then
			cnf_addReg <= (others => '0');
		
		elsif (cnf_addIncEn='1') then
			cnf_addReg <= std_logic_vector(unsigned(cnf_addReg) + 1);

		elsif (cnf_addDataSel='1' and cnf_en='0' and cnf_shiftEn='1') then
			cnf_addReg <= cnf_addReg(11 downto 0) & cnf_dIn;

		end if;

	end if;

end process;


-- DOutShiftReg -------------------------------------------------------------------------------------------------
-- Data Out Shift Register, parallel loaded by FSM
----------------------------------------------------------------------------------------------------------------
DOutShiftReg : process(clk)
begin
	if (clk'event and clk='1') then

		if (nReset='0') then
			cnf_dOutReg <= (others => '0');
		
		elsif (cnf_dOutLoadEn='1') then
			cnf_dOutReg <= mem_dOutLU;

		elsif (cnf_en='0' and cnf_shiftEn='1') then
			cnf_dOutReg <= cnf_dOutReg(14 downto 0) & '0';

		end if;

	end if;

end process;

cnf_dOut <= cnf_dOutReg(15);

-- MemInputSelect ---------------------------------------------------------------------------------------------
-- Multiplexor to select memory inputs, either config or main input ports
----------------------------------------------------------------------------------------------------------------
MemInputSelect : process (cnf_en, cnf_dInReg, dIn, cnf_addReg, add, cnf_memWEn, wEn, cnf_memEn, enable)
begin

	if (cnf_en = '1') then
		mem_dIn	<= cnf_dInReg;
		mem_add	<= cnf_addReg;
		mem_wEn	<= cnf_memWEn;
		mem_en	<= cnf_memEn;
	
	else
		mem_dIn	<= dIn;
		mem_add	<= add;
		mem_wEn	<= wEn;
		mem_en	<= enable;

	end if;

end process;


-- Mem --------------------------------------------------------------------------------------------------------
-- Instantiation of the memory
----------------------------------------------------------------------------------------------------------------
Mem : Memory
port map (	clk				=> clk,
				enable			=> mem_en,
				nReset			=> nReset,
				wEn				=> mem_wEn,
				add				=> mem_add,
				dIn				=> mem_dIn,
				dOutLU			=> mem_dOutLU,
				dOutU				=> mem_dOutU);

dOutLU <= mem_dOutLU;
dOutU <= mem_dOutU;

end General;

