---------------------------------------------------------------------------------------------------------------
-- RISA Project
-- Author: A.Greensted
-- Module:
-- Description:
-- History:
---------------------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library CommonLib;
use CommonLib.CommonPkg.all;

library ConfigLib;
use ConfigLib.ConfigPkg.all;

entity IOBlock is
	port (	clk					: in		std_logic;
				nReset				: in		std_logic;

				-- Pad Connections
				padAIn				: in		std_logic;
				padAOut				: out		std_logic;
				padAOutEn			: out		std_logic;

				padBIn				: in		std_logic;
				padBOut				: out		std_logic;
				padBOutEn			: out		std_logic;

				-- Fabirc Connections
				linkCAIn				: in		std_logic;
				linkCBIn				: in		std_logic;
				linkCCIn				: in		std_logic;
				linkCDIn				: in		std_logic;

				linkR1In				: in		std_logic_vector(1 downto 0);
				linkR2In				: in		std_logic;
				linkR4In				: in		std_logic;

				linkPAOut			: out		std_logic;
				linkPBOut			: out		std_logic;

				linkIOOut			: out		std_logic_vector(4 downto 0);

				shChnIn				: in		std_logic;
				shChnOut				: out		std_logic;
				cyChnIn				: in		std_logic;
				cyChnOut				: out		std_logic;

				snapDIn				: in		std_logic;
				snapDOut				: out		std_logic;

				-- Routing Configuration
				rSelChnDIn			: in		std_logic;
				rSelChnDOut			: out		std_logic;
				rSelChnShiftEn		: in		std_logic;

				rCnfChnDIn			: in		std_logic;
				rCnfChnDOut			: out		std_logic;
				rCnfChnShiftEn		: in		std_logic;
				rCnfChnLoadEn		: in		std_logic;
				rCnfChnReadBackEn	: in		std_logic);
end IOBlock;

architecture General of IOBlock is

	signal pAEn						: std_logic;
	signal pAOut					: std_logic;
	signal pAIn						: std_logic;
	signal pAInSynched			: std_logic;
	signal pA						: std_logic;
	signal pAInPulse				: std_logic;
	signal pAInLatched			: std_logic;
	signal pAInLatchEn			: std_logic;
	signal latchAEn				: std_logic;
	signal aOut						: std_logic;

	signal pBEn						: std_logic;
	signal pBOut					: std_logic;
	signal pBIn						: std_logic;
	signal pBInSynched			: std_logic;
	signal pB						: std_logic;
	signal pBInPulse				: std_logic;
	signal pBInLatched			: std_logic;
	signal pBInLatchEn			: std_logic;
	signal latchBEn				: std_logic;
	signal bOut						: std_logic;

	-- {{{ Function: selectPadOutput
	impure function selectPadOutput( connectionNum : std_logic_vector(3 downto 0) ) return std_logic is
		variable data : std_logic;
	begin
		-- Assign default value
		data := '0';

		case connectionNum is
			when b"0000"	=> data := '0';			-- 0

			when b"0001"	=> data := cyChnIn;		-- 1
			when b"0010"	=> data := shChnIn;		-- 2

			when b"0011"	=> data := snapDIn;		-- 3

			when b"0100"	=> data := linkCAIn;		-- 4
			when b"0101"	=> data := linkCBIn;		-- 5
			when b"0110"	=> data := linkCCIn;		-- 6
			when b"0111"	=> data := linkCDIn;		-- 7
			when b"1000"	=> data := linkR1In(0);	-- 8
			when b"1001"	=> data := linkR1In(1);	-- 9
			when b"1010"	=> data := linkR2In;		-- 10
			when b"1011"	=> data := linkR4In;		-- 11

			when others		=> data := '0';
		end case;

		return data;

	end function; -- }}}

	-- {{{ Function: selectSnapOutput
	impure function selectSnapOutput( connectionNum : std_logic_vector(3 downto 0) ) return std_logic is
		variable data : std_logic;
	begin
		-- Assign default value
		data := '0';

		case connectionNum is
			when b"0000"	=> data := cyChnIn;		-- 0
			when b"0001"	=> data := shChnIn;		-- 1

			when b"0010"	=> data := linkCAIn;		-- 2
			when b"0011"	=> data := linkCBIn;		-- 3
			when b"0100"	=> data := linkCCIn;		-- 4
			when b"0101"	=> data := linkCDIn;		-- 5
			when b"0110"	=> data := linkR1In(0);	-- 6
			when b"0111"	=> data := linkR1In(1);	-- 7
			when b"1000"	=> data := linkR2In;		-- 8
			when b"1001"	=> data := linkR4In;		-- 9

			when b"1010"	=> data := aOut;			-- 10
			when b"1011"	=> data := bOut;			-- 11

			when others		=> data := '0';
		end case;

		return data;

	end function; -- }}}

	-- {{{ Function: selectFabricOutput
	impure function selectFabricOutput( connectionNum : std_logic_vector(1 downto 0) ) return std_logic is
		variable data : std_logic;
	begin
		-- Assign default value
		data := '0';

		case connectionNum is
			when b"00"	=> data := '0';		-- 0
			when b"01"	=> data := snapDIn;	-- 1
			when b"10"	=> data := aOut;		-- 2
			when b"11"	=> data := bOut;		-- 3

			when others	=> data := '0';
		end case;

		return data;

	end function; -- }}}

	signal rCnfChn					: std_logic_vector(58 downto 0);
	signal rCnfBit					: std_logic_vector(57 downto 0);

	signal rCnfChnDInBP			: std_logic;
	signal rCnfChnDOutBP			: std_logic;
	signal rCnfChnShiftEnBP		: std_logic;
	signal rCnfChnLoadEnBP		: std_logic;
	signal rCnfChnReadBackEnBP	: std_logic;

begin

IOBlockRouteCnfBypassUnit : ConfigSelectUnit
port map (	clk						=> clk,
				nReset					=> nReset,
				selChnDIn				=> rSelChnDIn,
				selChnDOut				=> rSelChnDOut,
				selChnShiftEn			=> rSelChnShiftEn,

				cnfChnDIn				=>	rCnfChnDIn,
				cnfChnDOut				=> rCnfChnDOut,
				cnfChnShiftEn			=> rCnfChnShiftEn,
				cnfChnLoadEn			=> rCnfChnLoadEn,
				cnfChnReadBackEn		=> rCnfChnReadBackEn,

				cnfChnDInBP				=> rCnfChnDInBP,
				cnfChnDOutBP			=> rCnfChnDOutBP,
				cnfChnShiftEnBP		=> rCnfChnShiftEnBP,
				cnfChnLoadEnBP			=> rCnfChnLoadEnBP,
				cnfChnReadBackEnBP	=> rCnfChnReadBackEnBP);

-- Connect Cnf Chain to Select Unit
rCnfChn(0)		<= rCnfChnDOutBP;
rCnfChnDInBP	<= rCnfChn(58);

CnfBitGeneration : for b in 0 to 57 generate
begin
	IOBlockCnfBit : ConfigurableBit
	port map (	clk					=> clk,
					nReset				=> nReset,
					cnfBit				=> rCnfBit(b),
					cnfChnDIn			=> rCnfChn(b),
					cnfChnDOut			=> rCnfChn(b+1),
					cnfChnShiftEn		=> rCnfChnShiftEnBP,
					cnfChnLoadEn		=> rCnfChnLoadEnBP,
					cnfChnReadBackEn	=> rCnfChnReadBackEnBP);
end generate;

-- Direct (unswitched) pad outputs
linkPAOut		<= aOut;
linkPBOut		<= bOut;

SwitchBox : process (aOut, bOut, cyChnIn, shChnIn,
							linkCAIn, linkCBIn, linkCCIn, linkCDIn,
							linkR1In, linkR2In, linkR4In, snapDIn,
							rCnfBit)
begin

	pAEn				<= selectPadOutput(		rCnfBit(3  downto  0));
	pAOut				<= selectPadOutput(		rCnfBit(7  downto  4));
	pAInLatchEn		<= selectPadOutput(		rCnfBit(11 downto  8));

	pBEn				<= selectPadOutput(		rCnfBit(15 downto 12));
	pBOut				<= selectPadOutput(		rCnfBit(19 downto 16));
	pBInLatchEn		<= selectPadOutput(		rCnfBit(23 downto 20));

	shChnOut			<= selectFabricOutput(	rCnfBit(25 downto 24));
	cyChnOut			<= selectFabricOutput(	rCnfBit(27 downto 26));

	linkIOOut(0)	<= selectFabricOutput(	rCnfBit(29 downto 28));		-- Connects to Cluster's CA input
	linkIOOut(1)	<= selectFabricOutput(	rCnfBit(31 downto 30));		-- Connects to Cluster's CB input
	linkIOOut(2)	<= selectFabricOutput(	rCnfBit(33 downto 32));		-- Connects to Cluster's CC input
	linkIOOut(3)	<= selectFabricOutput(	rCnfBit(35 downto 34));		-- Connects to Cluster's CD input
	linkIOOut(4)	<= selectFabricOutput(	rCnfBit(37 downto 36));		-- Connects to Cluster's R4 input

	snapDOut			<= selectSnapOutput(		rCnfBit(41 downto 38));

end process;

--with rCnfBit(3 downto 0) select pAEn <= '0' when b"0000", cyChnIn when b"0001", shChnIn when b"0010", snapDIn when b"0011", linkCAIn when b"0100", linkCBIn when b"0101", linkCCIn when b"0110", linkCDIn when b"0111", linkR1In(0) when b"1000", linkR1In(1) when b"1001", linkR2In when b"1010", linkR4In when b"1011", '0' when others;

-- Pad A Config ---------------------------------------------------------------------------------------------
-- Configurable signals for Pad A
-------------------------------------------------------------------------------------------------------------
padAOutEn	<= not pAEn				when (rCnfBit(42)='1') else pAEn;
padAOut		<= not pAOut			when (rCnfBit(43)='1') else pAOut;
pAIn			<= not padAIn			when (rCnfBit(44)='1') else padAIn;

with rCnfBit(46 downto 45) select aOut <=	pAInSynched	when b"00",
														pAInPulse	when b"01",
														pA				when b"10",
														pAInLatched	when b"11",
														'0'			when others;

latchAEn		<= not pAInLatchEn	when (rCnfBit(47)='1') else pAInLatchEn;


-- Pad B Config ---------------------------------------------------------------------------------------------
-- Configurable signals for Pad B
-------------------------------------------------------------------------------------------------------------
padBOutEn	<= not pBEn				when (rCnfBit(48)='1') else pBEn;
padBOut		<= not pBOut			when (rCnfBit(49)='1') else pBOut;
pBIn			<= not padBIn			when (rCnfBit(50)='1') else padBIn;

with rCnfBit(52 downto 51) select bOut <=	pBInSynched	when b"00",
														pBInPulse	when b"01",
														pB				when b"10",
														pBInLatched	when b"11",
														'0'			when others;

latchBEn		<= not pBInLatchEn	when (rCnfBit(53)='1') else pBInLatchEn;


-- Pad ------------------------------------------------------------------------------------------------------
-- Create Pad input logic
-------------------------------------------------------------------------------------------------------------
SYNC_ASync : Synchroniser
port map (	clk	=> clk,
				dIn	=> pAIn,
				dOut	=> pAInSynched);

SYNC_BSync : Synchroniser
port map (	clk	=> clk,
				dIn	=> pBIn,
				dOut	=> pBInSynched);


Latches : process (clk, nReset)
--	variable
begin

	if (clk'event and clk='1') then
		if (nReset='0') then
			pA				<= '0';
			pB				<= '0';
			pAInLatched <= '0';
			pBInLatched <= '0';

		else
			if (latchAEn='1') then
				pA				<= pAInSynched;
				pAInLatched	<= pAIn;
			end if;

			if (latchBEn='1') then
				pB				<= pBInSynched;
				pBInLatched	<= pBIn;
			end if;


		end if;

	end if;

end process;

pAInPulse <= pAInSynched and not pA;
pBInPulse <= pBInSynched and not pB;

end General;
