---------------------------------------------------------------------------------------------------------------
-- RISA Project
-- Author: A.Greensted
-- Module:
-- Description:
-- History:
---------------------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

library SnapLib;
use SnapLib.SnapPkg.all;

library FPGALib;
use FPGALib.FPGAPkg.all;

library ConfigLib;
use ConfigLib.ConfigPkg.all;

library CommonLib;
use CommonLib.CommonPkg.all;

entity RISA is
	port (	clk					: in		std_logic;
				nReset				: in		std_logic;

				-- Config Access Port
				cap_clk				: in		std_logic;
				cap_nReset			: in		std_logic;
				cap_enable			: in		std_logic;
				cap_ctrlDataSel	: in		std_logic;
				cap_shiftExecSel	: in		std_logic;
				cap_dIn				: in		std_logic;
				cap_dOut				: out		std_logic;
				cap_busy				: out		std_logic;

				-- FPGA: IO Connections
				fIO_N_in				: in		std_logic_vector(11 downto 0);
				fIO_N_out			: out		std_logic_vector(11 downto 0);
				fIO_N_outEn			: out		std_logic_vector(11 downto 0);
				fIO_S_in				: in		std_logic_vector(11 downto 0);
				fIO_S_out			: out		std_logic_vector(11 downto 0);
				fIO_S_outEn			: out		std_logic_vector(11 downto 0);
				fIO_E_in				: in		std_logic_vector(11 downto 0);
				fIO_E_out			: out		std_logic_vector(11 downto 0);
				fIO_E_outEn			: out		std_logic_vector(11 downto 0);
				fIO_W_in				: in		std_logic_vector(11 downto 0);
				fIO_W_out			: out		std_logic_vector(11 downto 0);
				fIO_W_outEn			: out		std_logic_vector(11 downto 0);

				-- FPGA: Gloabl IO Connections
				fGIn					: in		std_logic_vector(1 downto 0);

				-- Snap: Snap Links
				sLnk_N_txOut		: out		std_logic;
				sLnk_N_rxIn			: in		std_logic;
				sLnk_N_CTSOut		: out		std_logic;
				sLnk_N_CTSIn		: in		std_logic;
				sLnk_E_txOut		: out		std_logic;
				sLnk_E_rxIn			: in		std_logic;
				sLnk_E_CTSOut		: out		std_logic;
				sLnk_E_CTSIn		: in		std_logic;
				sLnk_S_txOut		: out		std_logic;
				sLnk_S_rxIn			: in		std_logic;
				sLnk_S_CTSOut		: out		std_logic;
				sLnk_S_CTSIn		: in		std_logic;
				sLnk_W_txOut		: out		std_logic;
				sLnk_W_rxIn			: in		std_logic;
				sLnk_W_CTSOut		: out		std_logic;
				sLnk_W_CTSIn		: in		std_logic;

				-- Snap: System UART
				sysUART_txOut		: out		std_logic;
				sysUART_rxIn		: in		std_logic;

				-- Snap: IO Ports
				sIO_N_in				: in		std_logic_vector(7 downto 0);
				sIO_N_out			: out		std_logic_vector(7 downto 0);
				sIO_N_outEn			: out		std_logic_vector(7 downto 0);
				sIO_E_in				: in		std_logic_vector(7 downto 0);
				sIO_E_out			: out		std_logic_vector(7 downto 0);
				sIO_E_outEn			: out		std_logic_vector(7 downto 0);
				sIO_W_in				: in		std_logic_vector(7 downto 0);
				sIO_W_out			: out		std_logic_vector(7 downto 0);
				sIO_W_outEn			: out		std_logic_vector(7 downto 0);

				-- Snap: Counters
				cnt_0_extCountEn	: in		std_logic;
				cnt_0_match0Out	: out		std_logic;
				cnt_0_match1Out	: out		std_logic;
				cnt_1_extCountEn	: in		std_logic;
				cnt_1_match0Out	: out		std_logic;
				cnt_1_match1Out	: out		std_logic;

				-- Snap: Flags & IRQs
				extFlagsIn			: in		std_logic_vector(1 downto 0);
				extIRQsIn			: in		std_logic_vector(1 downto 0));

end RISA;

architecture General of RISA is

	signal nReset_sync							: std_logic;

	-- ConfigurableMemory
	signal mem_wEn									: std_logic;
	signal mem_add									: std_logic_vector(12 downto 0);
	signal mem_dIn									: std_logic_vector(15 downto 0);
	signal mem_dOutLU								: std_logic_vector(15 downto 0);
	signal mem_dOutU								: std_logic_vector(15 downto 0);

	-- Random Number Generator
	signal rng_dOut								: std_logic_vector(15 downto 0);

	-- Config Access Port
	signal cnf_CAP_to_Arbiter					: std_logic;
	signal cnf_Arbiter_to_CAP					: std_logic;
	signal cnf_shiftEn							: std_logic;
	signal cnf_ctrl								: std_logic_vector(1 downto 0);
	signal cnf_target								: CONFIGTYPE_CAP_TARGET;
	signal cnf_targetSetEn						: std_logic_vector(3 downto 0);
	
	signal start_CAP								: std_logic;
	signal release_CAP							: std_logic;
	signal controlData_CAP						: std_logic_vector(10 downto 0);
	signal sendData_CAP							: std_logic_vector(15 downto 0);
	signal returnData_CAP						: std_logic_vector(15 downto 0);

	-- Snap
	signal snap_nReset							: std_logic;
	signal snap_extFlagsIn						: std_logic_vector(7 downto 0);
	signal snap_extIRQsIn						: std_logic_vector(7 downto 0);

	signal extFlagsIn_sync						: std_logic_vector(1 downto 0);
	signal extIRQsIn_sync						: std_logic_vector(1 downto 0);

	signal sLnk_N_rxIn_sync						: std_logic;
	signal sLnk_N_CTSIn_sync					: std_logic;
	signal sLnk_S_rxIn_sync						: std_logic;
	signal sLnk_S_CTSIn_sync					: std_logic;
	signal sLnk_E_rxIn_sync						: std_logic;
	signal sLnk_E_CTSIn_sync					: std_logic;
	signal sLnk_W_rxIn_sync						: std_logic;
	signal sLnk_W_CTSIn_sync					: std_logic;

	signal sysUART_rxIn_sync					: std_logic;

	signal sIO_N_in_sync							: std_logic_vector(7 downto 0);
	signal sIO_E_in_sync							: std_logic_vector(7 downto 0);
	signal sIO_W_in_sync							: std_logic_vector(7 downto 0);

	signal cnt_0_extCountEn_sync				: std_logic;
	signal cnt_1_extCountEn_sync				: std_logic;

	-- FPGA
	signal fIO_N_in_sync							: std_logic_vector(11 downto 0);
	signal fIO_S_in_sync							: std_logic_vector(11 downto 0);
	signal fIO_E_in_sync							: std_logic_vector(11 downto 0);
	signal fIO_W_in_sync							: std_logic_vector(11 downto 0);

	signal fGIn_sync								: std_logic_vector(1 downto 0);

	signal fpga_snapDIn_N						: std_logic_vector(5 downto 0);
	signal fpga_snapDOut_N						: std_logic_vector(5 downto 0);
	signal fpga_snapDIn_S						: std_logic_vector(5 downto 0);
	signal fpga_snapDOut_S						: std_logic_vector(5 downto 0);
	signal fpga_snapDIn_E						: std_logic_vector(5 downto 0);
	signal fpga_snapDOut_E						: std_logic_vector(5 downto 0);
	signal fpga_snapDIn_W						: std_logic_vector(5 downto 0);
	signal fpga_snapDOut_W						: std_logic_vector(5 downto 0);

	-- ConfigArbiter
	signal uP_logicCnfChain_lockRequest		: std_logic;
	signal uP_logicCnfChain_lock				: std_logic;
	signal uP_logicCnfChain_shiftEn			: std_logic;
	signal uP_logicCnfChain_loadEn			: std_logic;
	signal uP_logicCnfChain_readBackEn		: std_logic;
	signal uP_logicCnfChain_dIn				: std_logic;
	signal uP_logicCnfChain_dOut				: std_logic;
	signal uP_logicSelChain_lockRequest		: std_logic;
	signal uP_logicSelChain_lock				: std_logic;
	signal uP_logicSelChain_shiftEn			: std_logic;
	signal uP_logicSelChain_feedbackEn		: std_logic;
	signal uP_logicSelChain_dIn				: std_logic;
	signal uP_logicSelChain_dOut				: std_logic;
	signal uP_routeCnfChain_lockRequest		: std_logic;
	signal uP_routeCnfChain_lock				: std_logic;
	signal uP_routeCnfChain_shiftEn			: std_logic;
	signal uP_routeCnfChain_loadEn			: std_logic;
	signal uP_routeCnfChain_readBackEn		: std_logic;
	signal uP_routeCnfChain_dIn				: std_logic;
	signal uP_routeCnfChain_dOut				: std_logic;
	signal uP_routeSelChain_lockRequest		: std_logic;
	signal uP_routeSelChain_lock				: std_logic;
	signal uP_routeSelChain_shiftEn			: std_logic;
	signal uP_routeSelChain_feedbackEn		: std_logic;
	signal uP_routeSelChain_dIn				: std_logic;
	signal uP_routeSelChain_dOut				: std_logic;

	signal uP_control_nReset					: std_logic;
	signal uP_control_enable					: std_logic;

	signal fpga_logicCnfChain_dIn				: std_logic;
	signal fpga_logicCnfChain_dOut			: std_logic;
	signal fpga_logicCnfChain_shiftEn		: std_logic;
	signal fpga_logicCnfChain_loadEn			: std_logic;
	signal fpga_logicCnfChain_readBackEn	: std_logic;
	signal fpga_logicSelChain_dIn				: std_logic;
	signal fpga_logicSelChain_dOut			: std_logic;
	signal fpga_logicSelChain_shiftEn		: std_logic;

	signal fpga_routeCnfChain_dIn				: std_logic;
	signal fpga_routeCnfChain_dOut			: std_logic;
	signal fpga_routeCnfChain_shiftEn		: std_logic;
	signal fpga_routeCnfChain_loadEn			: std_logic;
	signal fpga_routeCnfChain_readBackEn	: std_logic;
	signal fpga_routeSelChain_dIn				: std_logic;
	signal fpga_routeSelChain_dOut			: std_logic;
	signal fpga_routeSelChain_shiftEn		: std_logic;

	signal uP_memCnf_dIn							: std_logic;
	signal uP_memCnf_dOut						: std_logic;
	signal uP_memCnf_shiftEn					: std_logic;
	signal uP_memCnf_addDataSel				: std_logic;
	signal uP_memCnf_readEn						: std_logic;
	signal uP_memCnf_writeEn					: std_logic;

begin

-- Input Synchronisers ----------------------------------------------------------------------------------------
-- Connect input synchronisers to the signals that require them
---------------------------------------------------------------------------------------------------------------
SYNC_nReset				: Synchroniser port map (	clk => clk,		dIn => nReset,				dOut => nReset_sync);

-- FIO
SyncFIOGen : for s in 0 to 11 generate
	SYNC_FIO_N_in		: Synchroniser port map (	clk => clk,		dIn => fIO_N_in(s),		dOut => fIO_N_in_sync(s));
	SYNC_FIO_S_in		: Synchroniser port map (	clk => clk,		dIn => fIO_S_in(s),		dOut => fIO_S_in_sync(s));
	SYNC_FIO_E_in		: Synchroniser port map (	clk => clk,		dIn => fIO_E_in(s),		dOut => fIO_E_in_sync(s));
	SYNC_FIO_W_in		: Synchroniser port map (	clk => clk,		dIn => fIO_W_in(s),		dOut => fIO_W_in_sync(s));
End generate;

-- FGIn
SyncFGInGen : for s in 0 to 1 generate
	SYNC_FGIn			: Synchroniser port map (	clk => clk,		dIn => fGIn(s),			dOut => fGIn_sync(s));
end generate;

-- SLnk
SYNC_SLnk_N_RxIn		: Synchroniser port map (	clk => clk,		dIn => sLnk_N_rxIn,		dOut => sLnk_N_rxIn_sync);
SYNC_SLnk_N_CTSIn		: Synchroniser port map (	clk => clk,		dIn => sLnk_N_CTSIn,		dOut => sLnk_N_CTSIn_sync);
SYNC_SLnk_S_RxIn		: Synchroniser port map (	clk => clk,		dIn => sLnk_S_rxIn,		dOut => sLnk_S_rxIn_sync);
SYNC_SLnk_S_CTSIn		: Synchroniser port map (	clk => clk,		dIn => sLnk_S_CTSIn,		dOut => sLnk_S_CTSIn_sync);
SYNC_SLnk_E_RxIn		: Synchroniser port map (	clk => clk,		dIn => sLnk_E_rxIn,		dOut => sLnk_E_rxIn_sync);
SYNC_SLnk_E_CTSIn		: Synchroniser port map (	clk => clk,		dIn => sLnk_E_CTSIn,		dOut => sLnk_E_CTSIn_sync);
SYNC_SLnk_W_RxIn		: Synchroniser port map (	clk => clk,		dIn => sLnk_W_rxIn,		dOut => sLnk_W_rxIn_sync);
SYNC_SLnk_W_CTSIn		: Synchroniser port map (	clk => clk,		dIn => sLnk_W_CTSIn,		dOut => sLnk_W_CTSIn_sync);

-- SysUART
SYNC_SysUART_rxIn		: Synchroniser port map (	clk => clk,		dIn => sysUART_rxIn,		dOut => sysUART_rxIn_sync);

-- SIO
SyncSIOGen : for s in 0 to 7 generate
	SYNC_SIO_N			: Synchroniser port map (	clk => clk,		dIn => sIO_N_in(s),		dOut => sIO_N_in_sync(s));
	SYNC_SIO_E			: Synchroniser port map (	clk => clk,		dIn => sIO_E_in(s),		dOut => sIO_E_in_sync(s));
	SYNC_SIO_W			: Synchroniser port map (	clk => clk,		dIn => sIO_W_in(s),		dOut => sIO_W_in_sync(s));
end generate;

-- Ext Flags/IRQs
SyncExtGen : for s in 0 to 1 generate
	SYNC_ExtFlagsIn	: Synchroniser port map (	clk => clk,		dIn => extFlagsIn(s),	dOut => extFlagsIn_sync(s));
	SYNC_ExtIRQsIn		: Synchroniser port map (	clk => clk,		dIn => extIRQsIn(s),		dOut => extIRQsIn_sync(s));
end generate;

SYNC_Cnt_0_extEn		: Synchroniser port map (	clk => clk,		dIn => cnt_0_extCountEn,	dOut => cnt_0_extCountEn_sync);
SYNC_Cnt_1_extEn		: Synchroniser port map (	clk => clk,		dIn => cnt_1_extCountEn,	dOut => cnt_1_extCountEn_sync);

-- RNG --------------------------------------------------------------------------------------------------------
-- A pseudo random numner source for the snap processor cores
---------------------------------------------------------------------------------------------------------------
RNG : RandomNumberGenerator
port map (	clk						=> clk,
				enable					=> '1',
				nReset					=> nReset_sync,
				dOut						=> rng_dOut);


-- CAP --------------------------------------------------------------------------------------------------------
-- The chip's configuration access port. All configuration is done through this module
---------------------------------------------------------------------------------------------------------------
CAP_CAP : ConfigAccessPort_CAP
port map (	cap_clk				=> cap_clk,
				cap_nReset			=> cap_nReset,
				cap_enable			=> cap_enable,
				cap_ctrlDataSel	=> cap_ctrlDataSel,
				cap_shiftExecSel	=> cap_shiftExecSel,
				cap_dIn				=> cap_dIn,
				cap_dOut				=> cap_dOut,
				cap_busy				=> cap_busy,
				
				start					=> start_CAP,
				release				=> release_CAP,
				controlData			=> controlData_CAP,
				sendData				=> sendData_CAP,
				returnData			=> returnData_CAP);

CAP_CNF : ConfigAccessPort_CNF
port map (	clk					=> clk,
				cnf_nReset			=> nReset_sync,
				cnf_dIn				=> cnf_Arbiter_to_CAP,
				cnf_dOut				=> cnf_CAP_to_Arbiter,
				cnf_shiftEn			=> cnf_shiftEn,
				cnf_ctrl				=> cnf_ctrl,
				cnf_target			=> cnf_target,
				cnf_targetSetEn	=> cnf_targetSetEn,

				start					=> start_CAP,
				release				=> release_CAP,
				controlData			=> controlData_CAP,
				sendData				=> sendData_CAP,
				returnData			=> returnData_CAP);


-- FPGAFabric -------------------------------------------------------------------------------------------------
-- The cells FPGA Fabric
---------------------------------------------------------------------------------------------------------------
Fabric : FPGA
generic map(NUM_COLS				=> 6,		-- X Dimension
				NUM_ROWS				=> 6)		-- Y Dimension

port map	(	clk					=> clk,
				nReset				=> nReset_sync,

				-- Pads
				padIn_N				=> fIO_N_in_sync,
				padOut_N				=> fIO_N_out,
				padOutEn_N			=> fIO_N_outEn,

				padIn_S				=> fIO_S_in_sync,
				padOut_S				=> fIO_S_out,
				padOutEn_S			=> fIO_S_outEn,

				padIn_E				=> fIO_E_in_sync,
				padOut_E				=> fIO_E_out,
				padOutEn_E			=> fIO_E_outEn,

				padIn_W				=> fIO_W_in_sync,
				padOut_W				=> fIO_W_out,
				padOutEn_W			=> fIO_W_outEn,

				-- Global Inputs
				globalInputs		=> fGIn_sync,

				-- Snap Connections (Snap connections to IO Blocks)
				snapDIn_N			=> fpga_snapDIn_N,
				snapDOut_N			=> fpga_snapDOut_N,
				snapDIn_S			=> fpga_snapDIn_S,
				snapDOut_S			=> fpga_snapDOut_S,
				snapDIn_E			=> fpga_snapDIn_E,
				snapDOut_E			=> fpga_snapDOut_E,
				snapDIn_W			=> fpga_snapDIn_W,
				snapDOut_W			=> fpga_snapDOut_W,

				-- Logic Configuration
				lSelChnDIn			=> fpga_logicSelChain_dOut,
				lSelChnDOut			=> fpga_logicSelChain_dIn,
				lSelChnShiftEn		=> fpga_logicSelChain_shiftEn,

				lCnfChnDIn			=> fpga_logicCnfChain_dOut,
				lCnfChnDOut			=> fpga_logicCnfChain_dIn,
				lCnfChnShiftEn		=> fpga_logicCnfChain_shiftEn,
				lCnfChnLoadEn		=> fpga_logicCnfChain_loadEn,
				lCnfChnReadBackEn	=> fpga_logicCnfChain_readBackEn,

				-- Routing Configuration
				rSelChnDIn			=> fpga_routeSelChain_dOut,
				rSelChnDOut			=> fpga_routeSelChain_dIn,
				rSelChnShiftEn		=> fpga_routeSelChain_shiftEn,

				rCnfChnDIn			=> fpga_routeCnfChain_dOut,
				rCnfChnDOut			=> fpga_routeCnfChain_dIn,
				rCnfChnShiftEn		=> fpga_routeCnfChain_shiftEn,
				rCnfChnLoadEn		=> fpga_routeCnfChain_loadEn,
				rCnfChnReadBackEn	=> fpga_routeCnfChain_readBackEn);


-- SnapCore ---------------------------------------------------------------------------------------------------
-- The cells snap processor
---------------------------------------------------------------------------------------------------------------
snap_nReset <= '0' when (nReset_sync='0' or uP_control_nReset='0') else '1';

snap_extFlagsIn <= fpga_snapDOut_W(3) & fpga_snapDOut_W(1) & fpga_snapDOut_S(3) & fpga_snapDOut_S(1) &
						 fpga_snapDOut_E(2) & fpga_snapDOut_E(4) & extFlagsIn_sync;

snap_extIRQsIn <=	fpga_snapDOut_W(4) & fpga_snapDOut_W(2) & fpga_snapDOut_S(4) & fpga_snapDOut_S(2) &
						fpga_snapDOut_E(1) & fpga_snapDOut_E(3) & extIRQsIn_sync;
SnapCore : Snap
port map (	clk									=> clk,
				enable								=> uP_control_enable,
				nReset								=> snap_nReset,
				mem_wEn								=> mem_wEn,
				mem_add								=> mem_add,
				mem_dIn								=> mem_dIn,
				mem_dOutLU							=> mem_dOutLU,
				mem_dOutU							=> mem_dOutU,

				-- Flags Registers
				extFlagsIn							=> snap_extFlagsIn,

				-- Interrupt Controller
				extIRQsIn							=> snap_extIRQsIn,

				-- Snap Links
				sLnk_N_txOut						=> sLnk_N_txOut,
				sLnk_N_rxIn							=> sLnk_N_rxIn_sync,
				sLnk_N_CTSOut						=> sLnk_N_CTSOut,
				sLnk_N_CTSIn						=> sLnk_N_CTSIn_sync,

				sLnk_E_txOut						=> sLnk_E_txOut,
				sLnk_E_rxIn							=> sLnk_E_rxIn_sync,
				sLnk_E_CTSOut						=> sLnk_E_CTSOut,
				sLnk_E_CTSIn						=> sLnk_E_CTSIn_sync,

				sLnk_S_txOut						=> sLnk_S_txOut,
				sLnk_S_rxIn							=> sLnk_S_rxIn_sync,
				sLnk_S_CTSOut						=> sLnk_S_CTSOut,
				sLnk_S_CTSIn						=> sLnk_S_CTSIn_sync,

				sLnk_W_txOut						=> sLnk_W_txOut,
				sLnk_W_rxIn							=> sLnk_W_rxIn_sync,
				sLnk_W_CTSOut						=> sLnk_W_CTSOut,
				sLnk_W_CTSIn						=> sLnk_W_CTSIn_sync,

				-- Counter 0
				cnt_0_extCountEn					=> cnt_0_extCountEn_sync,
				cnt_0_match0Out					=> cnt_0_match0Out,
				cnt_0_match1Out					=> cnt_0_match1Out,

				-- Counter 1
				cnt_1_extCountEn					=> cnt_1_extCountEn_sync,
				cnt_1_match0Out					=> cnt_1_match0Out,
				cnt_1_match1Out					=> cnt_1_match1Out,

				-- System UART
				sysUART_txOut						=> sysUART_txOut,
				sysUART_rxIn						=> sysUART_rxIn_sync,

				-- Randon Number Generator
				rngDIn								=> rng_dOut,

				-- IO Ports
				ioP_N_dIn							=> sIO_N_in_sync,
				ioP_N_outEn							=> sIO_N_outEn,
				ioP_N_dOut							=> sIO_N_out,

				ioP_E_dIn							=> sIO_E_in_sync,
				ioP_E_outEn							=> sIO_E_outEn,
				ioP_E_dOut							=> sIO_E_out,

				ioP_W_dIn							=> sIO_W_in_sync,
				ioP_W_outEn							=> sIO_W_outEn,
				ioP_W_dOut							=> sIO_W_out,

				-- Fabric IO Ports
				fabIO_N_dIn							=> fpga_snapDOut_N,		-- Connect snapOut from FPGA fabric to Snap Fabric IO Port input
				fabIO_N_dOut						=> fpga_snapDIn_N,		-- Connect Snap Fabric IO Port output to the FPGA fabric snapIn input

				fabIO_S_dIn							=> fpga_snapDOut_S,
				fabIO_S_dOut						=> fpga_snapDIn_S,

				fabIO_E_dIn							=> fpga_snapDOut_E,
				fabIO_E_dOut						=> fpga_snapDIn_E,

				fabIO_W_dIn							=> fpga_snapDOut_W,
				fabIO_W_dOut						=> fpga_snapDIn_W,

				-- Fabric CAP Logic
				cap_L_cnfChainLockRequest		=> uP_logicCnfChain_lockRequest,
				cap_L_cnfChainLock				=> uP_logicCnfChain_lock,
				cap_L_cnfChainShiftEn			=> uP_logicCnfChain_shiftEn,
				cap_L_cnfChainLoadEn				=> uP_logicCnfChain_loadEn,
				cap_L_cnfChainReadBackEn		=> uP_logicCnfChain_readBackEn,
				cap_L_cnfChainOut					=> uP_logicCnfChain_dIn,
				cap_L_cnfChainIn					=> uP_logicCnfChain_dOut,

				cap_L_selChainLockRequest		=> uP_logicSelChain_lockRequest,
				cap_L_selChainLock				=> uP_logicSelChain_lock,
				cap_L_selChainShiftEn			=> uP_logicSelChain_shiftEn,
				cap_L_selChainFeedbackEn		=> uP_logicSelChain_feedbackEn,
				cap_L_selChainOut					=> uP_logicSelChain_dIn,
				cap_L_selChainIn					=> uP_logicSelChain_dOut,

				-- Fabric CAP Route
				cap_R_cnfChainLockRequest		=> uP_routeCnfChain_lockRequest,
				cap_R_cnfChainLock				=> uP_routeCnfChain_lock,
				cap_R_cnfChainShiftEn			=> uP_routeCnfChain_shiftEn,
				cap_R_cnfChainLoadEn				=> uP_routeCnfChain_loadEn,
				cap_R_cnfChainReadBackEn		=> uP_routeCnfChain_readBackEn,
				cap_R_cnfChainOut					=> uP_routeCnfChain_dIn,
				cap_R_cnfChainIn					=> uP_routeCnfChain_dOut,

				cap_R_selChainLockRequest		=> uP_routeSelChain_lockRequest,
				cap_R_selChainLock				=> uP_routeSelChain_lock,
				cap_R_selChainShiftEn			=> uP_routeSelChain_shiftEn,
				cap_R_selChainFeedbackEn		=> uP_routeSelChain_feedbackEn,
				cap_R_selChainOut					=> uP_routeSelChain_dIn,
				cap_R_selChainIn					=> uP_routeSelChain_dOut);


-- CnfMemory -------------------------------------------------------------------------------------------------
-- The Prorgam and data memory of the cell's snap processor
---------------------------------------------------------------------------------------------------------------
CnfMemory : ConfigurableMemory
port map (	clk									=> clk,
				enable								=>	'1',
				nReset								=> nReset_sync,

				wEn									=> mem_wEn,
				add									=>	mem_add,
				dIn									=>	mem_dIn,
				dOutLU								=> mem_dOutLU,
				dOutU									=> mem_dOutU,

				cnf_dIn								=> uP_memCnf_dOut,
				cnf_dOut								=> uP_memCnf_dIn,
				cnf_shiftEn							=> uP_memCnf_shiftEn,
				cnf_addDataSel						=> uP_memCnf_addDataSel,
				cnf_readEn							=> uP_memCnf_readEn,
				cnf_writeEn							=> uP_memCnf_writeEn);


-- CnfArbiter -------------------------------------------------------------------------------------------------
-- The cells Configuration Arbiter
---------------------------------------------------------------------------------------------------------------
CnfArbiter : ConfigArbiter
port map (	clk									=> clk,
				nReset								=> nReset_sync,

				cnf_dIn								=> cnf_CAP_to_Arbiter,
				cnf_dOut								=> cnf_Arbiter_to_CAP,
				cnf_shiftEn							=> cnf_shiftEn,
				cnf_ctrl								=> cnf_ctrl,
				cnf_target							=> cnf_target,
				cnf_targetSetEn					=> cnf_targetSetEn(0),

				-- Logic Configuration driving FPGA
				fpga_logicCnfChain_dIn			=> fpga_logicCnfChain_dIn,					-- ConfigArbiter input
				fpga_logicCnfChain_dOut			=> fpga_logicCnfChain_dOut,				-- ConfigArbiter output
				fpga_logicCnfChain_shiftEn		=> fpga_logicCnfChain_shiftEn,
				fpga_logicCnfChain_loadEn		=> fpga_logicCnfChain_loadEn,
				fpga_logicCnfChain_readBackEn	=> fpga_logicCnfChain_readBackEn,

				-- Logic Configuration from uP
				uP_logicCnfChain_dIn				=> uP_logicCnfChain_dIn,
				uP_logicCnfChain_dOut			=> uP_logicCnfChain_dOut,
				uP_logicCnfChain_shiftEn		=> uP_logicCnfChain_shiftEn,
				uP_logicCnfChain_loadEn			=> uP_logicCnfChain_loadEn,
				uP_logicCnfChain_readBackEn	=> uP_logicCnfChain_readBackEn,

				uP_logicCnfChain_lockRequest	=> uP_logicCnfChain_lockRequest,
				uP_logicCnfChain_lock			=> uP_logicCnfChain_lock,

				-- Logic Target Configuration driving fabric
				fpga_logicSelChain_dIn			=> fpga_logicSelChain_dIn,
				fpga_logicSelChain_dOut			=> fpga_logicSelChain_dOut,
				fpga_logicSelChain_shiftEn		=> fpga_logicSelChain_shiftEn,

				-- Logic Target Configuration from uP
				uP_logicSelChain_dIn				=> uP_logicSelChain_dIn,
				uP_logicSelChain_dOut			=> uP_logicSelChain_dOut,
				uP_logicSelChain_shiftEn		=> uP_logicSelChain_shiftEn,
				uP_logicSelChain_feedbackEn	=> uP_logicSelChain_feedbackEn,

				uP_logicSelChain_lockRequest	=> uP_logicSelChain_lockRequest,
				uP_logicSelChain_lock			=> uP_logicSelChain_lock,

				-- Route Configuration driving Fabric
				fpga_routeCnfChain_dIn			=> fpga_routeCnfChain_dIn,
				fpga_routeCnfChain_dOut			=> fpga_routeCnfChain_dOut,
				fpga_routeCnfChain_shiftEn		=> fpga_routeCnfChain_shiftEn,
				fpga_routeCnfChain_loadEn		=> fpga_routeCnfChain_loadEn,
				fpga_routeCnfChain_readBackEn	=> fpga_routeCnfChain_readBackEn,

				-- Route Configuration from uP
				uP_routeCnfChain_dIn				=> uP_routeCnfChain_dIn,
				uP_routeCnfChain_dOut			=> uP_routeCnfChain_dOut,
				uP_routeCnfChain_shiftEn		=> uP_routeCnfChain_shiftEn,
				uP_routeCnfChain_loadEn			=> uP_routeCnfChain_loadEn,
				uP_routeCnfChain_readBackEn	=> uP_routeCnfChain_readBackEn,

				uP_routeCnfChain_lockRequest	=> uP_routeCnfChain_lockRequest,
				uP_routeCnfChain_lock			=> uP_routeCnfChain_lock,

				-- Route Target Configuration driving fabric
				fpga_routeSelChain_dIn			=> fpga_routeSelChain_dIn,
				fpga_routeSelChain_dOut			=> fpga_routeSelChain_dOut,
				fpga_routeSelChain_shiftEn		=> fpga_routeSelChain_shiftEn,

				-- Route Target Configuration from uP
				uP_routeSelChain_dIn				=> uP_routeSelChain_dIn,
				uP_routeSelChain_dOut			=> uP_routeSelChain_dOut,
				uP_routeSelChain_shiftEn		=> uP_routeSelChain_shiftEn,
				uP_routeSelChain_feedbackEn	=> uP_routeSelChain_feedbackEn,

				uP_routeSelChain_lockRequest	=> uP_routeSelChain_lockRequest,
				uP_routeSelChain_lock			=> uP_routeSelChain_lock,

				-- Memory Configuration
				uP_memCnf_dIn						=> uP_memCnf_dIn,
				uP_memCnf_dOut						=> uP_memCnf_dOut,
				uP_memCnf_shiftEn					=> uP_memCnf_shiftEn,
				uP_memCnf_addDataSel				=> uP_memCnf_addDataSel,
				uP_memCnf_readEn					=> uP_memCnf_readEn,
				uP_memCnf_writeEn					=> uP_memCnf_writeEn,

				-- uP Control
				uP_control_nReset					=> uP_control_nReset,
				uP_control_enable					=> uP_control_enable);

end General;
