---------------------------------------------------------------------------------------------------------------
-- RISA Project
-- Author: A.Greensted
-- Module: RISA_Wrap
-- Description: Wrapper around RISA, for including IO Cells
-- History:
---------------------------------------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;

library RISALib;
use RISALib.RISAPkg.all;

entity RISA_Wrap is
	port (	clk					: in		std_logic;
				nReset				: in		std_logic;

				-- Config Access Port
				cap_clk				: in		std_logic;
				cap_nReset			: in		std_logic;
				cap_enable			: in		std_logic;
				cap_ctrlDataSel	: in		std_logic;
				cap_shiftExecSel	: in		std_logic;
				cap_dIn				: in		std_logic;
				cap_dOut				: out		std_logic;
				cap_busy				: out		std_logic;

				-- FPGA: IO Connections
				fIO_N					: inout	std_logic_vector(11 downto 0);
				fIO_S					: inout	std_logic_vector(11 downto 0);
				fIO_E					: inout	std_logic_vector(11 downto 0);
				fIO_W					: inout	std_logic_vector(11 downto 0);

				-- FPGA: Gloabl IO Connections
				fGIn					: in		std_logic_vector(1 downto 0);

				-- Snap: Snap Links
				sLnk_N_txOut		: out		std_logic;
				sLnk_N_rxIn			: in		std_logic;
				sLnk_N_CTSOut		: out		std_logic;
				sLnk_N_CTSIn		: in		std_logic;
				sLnk_E_txOut		: out		std_logic;
				sLnk_E_rxIn			: in		std_logic;
				sLnk_E_CTSOut		: out		std_logic;
				sLnk_E_CTSIn		: in		std_logic;
				sLnk_S_txOut		: out		std_logic;
				sLnk_S_rxIn			: in		std_logic;
				sLnk_S_CTSOut		: out		std_logic;
				sLnk_S_CTSIn		: in		std_logic;
				sLnk_W_txOut		: out		std_logic;
				sLnk_W_rxIn			: in		std_logic;
				sLnk_W_CTSOut		: out		std_logic;
				sLnk_W_CTSIn		: in		std_logic;

				-- Snap: System UART
				sysUART_txOut		: out		std_logic;
				sysUART_rxIn		: in		std_logic;

				-- Snap: IO Ports
				sIO_N					: inout	std_logic_vector(7 downto 0);
				sIO_E					: inout	std_logic_vector(7 downto 0);
				sIO_W					: inout	std_logic_vector(7 downto 0);

				-- Snap: Counters
				cnt_0_extCountEn	: in		std_logic;
				cnt_0_match0Out	: out		std_logic;
				cnt_0_match1Out	: out		std_logic;
				cnt_1_extCountEn	: in		std_logic;
				cnt_1_match0Out	: out		std_logic;
				cnt_1_match1Out	: out		std_logic;

				-- Snap: Flags & IRQs
				extFlagsIn			: in		std_logic_vector(1 downto 0);
				extIRQsIn			: in		std_logic_vector(1 downto 0));

end RISA_Wrap;

architecture General of RISA_Wrap is

	-- Input - Inline, CMOS, 3.3V, 4x drive, Schmitt
	component WC3I42
		port (	pad		: in		std_logic;
					dI			: out		std_logic);
	end component;

	-- Input - Inline, CMOS, 3.3V, 4x drive, Schmitt, Weak pull-up
	component WC3I42U
		port (	pad		: in		std_logic;
					dI			: out		std_logic);
	end component;

	-- Input - Inline, CMOS, 3.3V, 4x drive, Schmitt, Weak pull-down
	component WC3I42D
		port (	pad		: in		std_logic;
					dI			: out		std_logic);
	end component;

	-- Output - Inline, CMOS, 3.3V, 8mA, slew rate B (slower)
	component WC3O40B
		port (	dO			: in		std_logic;
					pad		: out		std_logic);
	end component;

	-- Bidirectional - Inline, CMOS, 3.3V, 4mA, Schmitt, Weak pull-up
	component WC3B42UB
		port (	dO			: in		std_logic;
					dI			: out		std_logic;
					en			: in		std_logic;
					pad		: inout	std_logic);
	end component;

	signal sig_clk						: std_logic;
	signal sig_nReset					: std_logic;

	-- Config Access Port
	signal sig_cap_clk				: std_logic;
	signal sig_cap_nReset			: std_logic;
	signal sig_cap_enable			: std_logic;
	signal sig_cap_ctrlDataSel		: std_logic;
	signal sig_cap_shiftExecSel	: std_logic;
	signal sig_cap_dIn				: std_logic;
	signal sig_cap_dOut				: std_logic;
	signal sig_cap_busy				: std_logic;

	-- FPGA: IO Connections
	signal sig_fIO_N_in				: std_logic_vector(11 downto 0);
	signal sig_fIO_N_out				: std_logic_vector(11 downto 0);
	signal sig_fIO_N_outEn			: std_logic_vector(11 downto 0);
	signal sig_fIO_S_in				: std_logic_vector(11 downto 0);
	signal sig_fIO_S_out				: std_logic_vector(11 downto 0);
	signal sig_fIO_S_outEn			: std_logic_vector(11 downto 0);
	signal sig_fIO_E_in				: std_logic_vector(11 downto 0);
	signal sig_fIO_E_out				: std_logic_vector(11 downto 0);
	signal sig_fIO_E_outEn			: std_logic_vector(11 downto 0);
	signal sig_fIO_W_in				: std_logic_vector(11 downto 0);
	signal sig_fIO_W_out				: std_logic_vector(11 downto 0);
	signal sig_fIO_W_outEn			: std_logic_vector(11 downto 0);

	-- FPGA: Gloabl IO Connections
	signal sig_fGIn					: std_logic_vector(1 downto 0);

	-- Snap: Snap Links
	signal sig_sLnk_N_txOut			: std_logic;
	signal sig_sLnk_N_rxIn			: std_logic;
	signal sig_sLnk_N_CTSOut		: std_logic;
	signal sig_sLnk_N_CTSIn			: std_logic;
	signal sig_sLnk_E_txOut			: std_logic;
	signal sig_sLnk_E_rxIn			: std_logic;
	signal sig_sLnk_E_CTSOut		: std_logic;
	signal sig_sLnk_E_CTSIn			: std_logic;
	signal sig_sLnk_S_txOut			: std_logic;
	signal sig_sLnk_S_rxIn			: std_logic;
	signal sig_sLnk_S_CTSOut		: std_logic;
	signal sig_sLnk_S_CTSIn			: std_logic;
	signal sig_sLnk_W_txOut			: std_logic;
	signal sig_sLnk_W_rxIn			: std_logic;
	signal sig_sLnk_W_CTSOut		: std_logic;
	signal sig_sLnk_W_CTSIn			: std_logic;

	-- Snap: System UART
	signal sig_sysUART_txOut		: std_logic;
	signal sig_sysUART_rxIn			: std_logic;

	-- Snap: IO Ports
	signal sig_sIO_N_in				: std_logic_vector(7 downto 0);
	signal sig_sIO_N_out				: std_logic_vector(7 downto 0);
	signal sig_sIO_N_outEn			: std_logic_vector(7 downto 0);
	signal sig_sIO_E_in				: std_logic_vector(7 downto 0);
	signal sig_sIO_E_out				: std_logic_vector(7 downto 0);
	signal sig_sIO_E_outEn			: std_logic_vector(7 downto 0);
	signal sig_sIO_W_in				: std_logic_vector(7 downto 0);
	signal sig_sIO_W_out				: std_logic_vector(7 downto 0);
	signal sig_sIO_W_outEn			: std_logic_vector(7 downto 0);

	-- Snap: Counters
	signal sig_cnt_0_extCountEn	: std_logic;
	signal sig_cnt_0_match0Out		: std_logic;
	signal sig_cnt_0_match1Out		: std_logic;
	signal sig_cnt_1_extCountEn	: std_logic;
	signal sig_cnt_1_match0Out		: std_logic;
	signal sig_cnt_1_match1Out		: std_logic;

	-- Snap: Flags & IRQs
	signal sig_extFlagsIn			: std_logic_vector(1 downto 0);
	signal sig_extIRQsIn				: std_logic_vector(1 downto 0);

begin

PAD_clk 					: WC3I42		port map (pad=>clk, 					dI=>sig_clk);						-- Input - schmitt
PAD_nReset				: WC3I42		port map (pad=>nReset, 				dI=>sig_nReset);					-- Input - schmitt

-- Config Access Port
PAD_cap_clk				: WC3I42		port map (pad=>cap_clk,				dI=>sig_cap_clk);
PAD_cap_nReset			: WC3I42D	port map (pad=>cap_nReset,			dI=>sig_cap_nReset);				-- Pull Down
PAD_cap_enable			: WC3I42		port map (pad=>cap_enable,			dI=>sig_cap_enable);
PAD_cap_ctrlDataSel	: WC3I42		port map (pad=>cap_ctrlDataSel,	dI=>sig_cap_ctrlDataSel);
PAD_cap_shiftExecSel	: WC3I42		port map (pad=>cap_shiftExecSel,	dI=>sig_cap_shiftExecSel);
PAD_cap_dIn				: WC3I42		port map (pad=>cap_dIn,				dI=>sig_cap_dIn);
PAD_cap_dOut			: WC3O40B	port map (pad=>cap_dOut,			dO=>sig_cap_dOut);
PAD_cap_busy			: WC3O40B	port map (pad=>cap_busy,			dO=>sig_cap_busy);

-- FPGA: IO Connections
PadFIOGen : for p in 0 to 11 generate
	PAD_fIO_N			: WC3B42UB	port map (pad=>fIO_N(p), dO=>sig_fIO_N_out(p), en=>sig_fIO_N_outEn(p), dI=>sig_fIO_N_in(p));
	PAD_fIO_S			: WC3B42UB	port map (pad=>fIO_S(p), dO=>sig_fIO_S_out(p), en=>sig_fIO_S_outEn(p), dI=>sig_fIO_S_in(p));
	PAD_fIO_E			: WC3B42UB	port map (pad=>fIO_E(p), dO=>sig_fIO_E_out(p), en=>sig_fIO_E_outEn(p), dI=>sig_fIO_E_in(p));
	PAD_fIO_W			: WC3B42UB	port map (pad=>fIO_W(p), dO=>sig_fIO_W_out(p), en=>sig_fIO_W_outEn(p), dI=>sig_fIO_W_in(p));
end generate;

-- FPGA: Gloabl IO Connections
PadFGInGen : for p in 0 to 1 generate
	PAD_fGIn				: WC3I42U	port map (pad=>fGIn(p),				dI=>sig_fGIn(p));					-- Pull Up
end generate;

-- Snap: Snap Links
PAD_sLnk_N_txOut		: WC3O40B	port map (pad=>sLnk_N_txOut,		dO=>sig_sLnk_N_txOut);
PAD_sLnk_N_rxIn		: WC3I42U	port map (pad=>sLnk_N_rxIn,		dI=>sig_sLnk_N_rxIn);			-- Pull Up (rxIn)
PAD_sLnk_N_CTSOut		: WC3O40B	port map (pad=>sLnk_N_CTSOut,		dO=>sig_sLnk_N_CTSOut);
PAD_sLnk_N_CTSIn		: WC3I42D	port map (pad=>sLnk_N_CTSIn,		dI=>sig_sLnk_N_CTSIn);			-- Pull Down (CTSIn)

PAD_sLnk_E_txOut		: WC3O40B	port map (pad=>sLnk_E_txOut,		dO=>sig_sLnk_E_txOut);
PAD_sLnk_E_rxIn 		: WC3I42U	port map (pad=>sLnk_E_rxIn,		dI=>sig_sLnk_E_rxIn);			-- Pull Up (rxIn)
PAD_sLnk_E_CTSOut		: WC3O40B	port map (pad=>sLnk_E_CTSOut,		dO=>sig_sLnk_E_CTSOut);
PAD_sLnk_E_CTSIn		: WC3I42D	port map (pad=>sLnk_E_CTSIn,		dI=>sig_sLnk_E_CTSIn);			-- Pull Down (CTSIn)

PAD_sLnk_S_txOut		: WC3O40B	port map (pad=>sLnk_S_txOut,		dO=>sig_sLnk_S_txOut);
PAD_sLnk_S_rxIn 		: WC3I42U	port map (pad=>sLnk_S_rxIn,		dI=>sig_sLnk_S_rxIn);			-- Pull Up (rxIn)
PAD_sLnk_S_CTSOut		: WC3O40B	port map (pad=>sLnk_S_CTSOut,		dO=>sig_sLnk_S_CTSOut);
PAD_sLnk_S_CTSIn		: WC3I42D	port map (pad=>sLnk_S_CTSIn,		dI=>sig_sLnk_S_CTSIn);			-- Pull Down (CTSIn)

PAD_sLnk_W_txOut		: WC3O40B	port map (pad=>sLnk_W_txOut,		dO=>sig_sLnk_W_txOut);
PAD_sLnk_W_rxIn	 	: WC3I42U	port map (pad=>sLnk_W_rxIn,		dI=>sig_sLnk_W_rxIn);			-- Pull Up (rxIn)
PAD_sLnk_W_CTSOut		: WC3O40B	port map (pad=>sLnk_W_CTSOut,		dO=>sig_sLnk_W_CTSOut);
PAD_sLnk_W_CTSIn		: WC3I42D	port map (pad=>sLnk_W_CTSIn,		dI=>sig_sLnk_W_CTSIn);			-- Pull Down (CTSIn)

-- Snap: System UART
PAD_sysUART_txOut		: WC3O40B	port map (pad=>sysUART_txOut,		dO=>sig_sysUART_txOut);
PAD_sysUART_rxIn  	: WC3I42U	port map (pad=>sysUART_rxIn,		dI=>sig_sysUART_rxIn);			-- Pull Up (rxIn)

-- Snap: IO Ports
PadSIOGen : for p in 0 to 7 generate
	PAD_sIO_N			: WC3B42UB	port map (pad=>sIO_N(p), dO=>sig_sIO_N_out(p), en=>sig_sIO_N_outEn(p), dI=>sig_sIO_N_in(p));
	PAD_sIO_E			: WC3B42UB	port map (pad=>sIO_E(p), dO=>sig_sIO_E_out(p), en=>sig_sIO_E_outEn(p), dI=>sig_sIO_E_in(p));
	PAD_sIO_W			: WC3B42UB	port map (pad=>sIO_W(p), dO=>sig_sIO_W_out(p), en=>sig_sIO_W_outEn(p), dI=>sig_sIO_W_in(p));
end generate;

-- Snap: Counters
PAD_cnt_0_extCountEn	: WC3I42D	port map (pad=>cnt_0_extCountEn,	dI=>sig_cnt_0_extCountEn);		-- Pull Down (extEn)
PAD_cnt_0_match0Out	: WC3O40B	port map (pad=>cnt_0_match0Out,	dO=>sig_cnt_0_match0Out);
PAD_cnt_0_match1Out	: WC3O40B	port map (pad=>cnt_0_match1Out,	dO=>sig_cnt_0_match1Out);

PAD_cnt_1_extCountEn	: WC3I42D	port map (pad=>cnt_1_extCountEn,	dI=>sig_cnt_1_extCountEn);		-- Pull Down (extEn)
PAD_cnt_1_match0Out	: WC3O40B	port map (pad=>cnt_1_match0Out,	dO=>sig_cnt_1_match0Out);
PAD_cnt_1_match1Out	: WC3O40B	port map (pad=>cnt_1_match1Out,	dO=>sig_cnt_1_match1Out);

-- Snap: Flags & IRQs
PadFlagGen : for p in 0 to 1 generate
	PAD_extFlagsIn		: WC3I42U	port map (pad=>extFlagsIn(p),		dI=>sig_extFlagsIn(p));
	PAD_extIRQsIn		: WC3I42U	port map (pad=>extIRQsIn(p),		dI=>sig_extIRQsIn(p));
end generate;

-- Instantiate the RISA Chip
RISAUnit : RISA
port map (	clk					=> sig_clk,
				nReset				=> sig_nReset,

				cap_clk				=> sig_cap_clk,
				cap_nReset			=> sig_cap_nReset,
				cap_enable			=> sig_cap_enable,
				cap_ctrlDataSel	=> sig_cap_ctrlDataSel,
				cap_shiftExecSel	=> sig_cap_shiftExecSel,
				cap_dIn				=> sig_cap_dIn,
				cap_dOut				=> sig_cap_dOut,
				cap_busy				=> sig_cap_busy,

				-- FPGA: IO Connections
				fIO_N_in				=> sig_fIO_N_in,
				fIO_N_out			=> sig_fIO_N_out,
				fIO_N_outEn			=> sig_fIO_N_outEn,
				fIO_S_in				=> sig_fIO_S_in,
				fIO_S_out			=> sig_fIO_S_out,
				fIO_S_outEn			=> sig_fIO_S_outEn,
				fIO_E_in				=> sig_fIO_E_in,
				fIO_E_out			=> sig_fIO_E_out,
				fIO_E_outEn			=> sig_fIO_E_outEn,
				fIO_W_in				=> sig_fIO_W_in,
				fIO_W_out			=> sig_fIO_W_out,
				fIO_W_outEn			=> sig_fIO_W_outEn,

				-- FPGA: Global IO Connections
				fGIn					=> sig_fGIn,

				-- Snap: Snap Links
				sLnk_N_txOut		=> sig_sLnk_N_txOut,
				sLnk_N_rxIn			=> sig_sLnk_N_rxIn,
				sLnk_N_CTSOut		=> sig_sLnk_N_CTSOut,
				sLnk_N_CTSIn		=> sig_sLnk_N_CTSIn,
				sLnk_E_txOut		=> sig_sLnk_E_txOut,
				sLnk_E_rxIn			=> sig_sLnk_E_rxIn,
				sLnk_E_CTSOut		=> sig_sLnk_E_CTSOut,
				sLnk_E_CTSIn		=> sig_sLnk_E_CTSIn,
				sLnk_S_txOut		=> sig_sLnk_S_txOut,
				sLnk_S_rxIn			=> sig_sLnk_S_rxIn,
				sLnk_S_CTSOut		=> sig_sLnk_S_CTSOut,
				sLnk_S_CTSIn		=> sig_sLnk_S_CTSIn,
				sLnk_W_txOut		=> sig_sLnk_W_txOut,
				sLnk_W_rxIn			=> sig_sLnk_W_rxIn,
				sLnk_W_CTSOut		=> sig_sLnk_W_CTSOut,
				sLnk_W_CTSIn		=> sig_sLnk_W_CTSIn,

				-- Snap: System UART
				sysUART_txOut		=> sig_sysUART_txOut,
				sysUART_rxIn		=> sig_sysUART_rxIn,

				-- Snap: IO Ports
				sIO_N_in				=> sig_sIO_N_in,
				sIO_N_out			=> sig_sIO_N_out,
				sIO_N_outEn			=> sig_sIO_N_outEn,
				sIO_E_in				=> sig_sIO_E_in,
				sIO_E_out			=> sig_sIO_E_out,
				sIO_E_outEn			=> sig_sIO_E_outEn,
				sIO_W_in				=> sig_sIO_W_in,
				sIO_W_out			=> sig_sIO_W_out,
				sIO_W_outEn			=> sig_sIO_W_outEn,

				-- Snap: Counters
				cnt_0_extCountEn	=> sig_cnt_0_extCountEn,
				cnt_0_match0Out	=> sig_cnt_0_match0Out,
				cnt_0_match1Out	=> sig_cnt_0_match1Out,
				cnt_1_extCountEn	=> sig_cnt_1_extCountEn,
				cnt_1_match0Out	=> sig_cnt_1_match0Out,
				cnt_1_match1Out	=> sig_cnt_1_match1Out,

				-- Snap: Flags & IRQs
				extFlagsIn			=> sig_extFlagsIn,
				extIRQsIn			=> sig_extIRQsIn);

end General;
